"""
An Experiment is a collection of queries, and manages their retrieved documents, agent answers and their respective
evaluations created by Evaluators.
For most use cases, this class should be the main entry point to interact with RAGElo.
"""

from __future__ import annotations

import csv
import json
import logging
import os
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Any, Literal, overload

import rich
from pydantic import BaseModel

from ragelo.logger import CLILogHandler, logger
from ragelo.types.evaluables import AgentAnswer, Document
from ragelo.types.query import Query
from ragelo.types.results import AnswerEvaluatorResult, EloTournamentResult, EvaluatorResult, RetrievalEvaluatorResult


class Experiment:
    """
    A class to manage and evaluate an experiment. An experiment is composed of multiple queries and their evaluations.
    This is the main class for interacting with RAGElo, and manages all the queries to which agents retrieve documents
      and answer questions.
    Attributes:
        experiment_name (str): An unique name for your experiment.
        queries: (dict[str, Query]): A dictionary with the queries for this experiment.
            Can also be accessed directly with __getitem__.
        elo_tournaments: (list[EloTournamentResult]): A list with the results for Elo Tournaments
            played over this set of queries.
    Methods:
        add_query(query, query_id, doc_id, metadata, force): Adds a query
            (either a Query object or a string) to the current experiment. Returns the id of the added query.
        add_retrieved_doc(doc, query_id, doc_id, score, agent, force): Adds a retrieved document to a query
        add_agent_answer(answer, agent, query_id, force): Adds an agent's answer to a query.
        add_evaluation(evaluation, should_save, force): Adds an evaluation to a retrieved document,
            an answer or a pairwise game.
        evaluate_retrieval(metrics, relevance_threshold): Evaluates the retrieval performance of the agents
            using the specified metrics and the evaluations generated by a `RetrievalEvaluator`.
            This requires the ir_measures package to be installed.
        get_qrels(relevance_key, relevance_threshold, output_path, output_format): Gets the qrels
            (dictionary of relevance) for the queries. Documents with a score above relevance_threshold are considered
            relevant. Optionally saves the result to output_path with output_format.
        get_runs(agents, output_path, output_format): Gets the runs (retrieval results) for the provided agents for
            all the queries. Optionally saves the result to output_path with output_format.
        save(): Saves the current state of the queries object to disk as a JSON object to `cache_path`.
        save_results(result): Saves an evaluation result to the `evaluations_cache_path` file.
        add_retrieved_docs_from_runfile(run_file_path, corpus, top_k): Adds retrieved documents from a trec-formatted
            run file to the queries. Will create the Document objects as needed.

        __len__(): Returns the number of queries.
        __iter__(): Returns an iterator over the queries.
        __getitem__(key: str) -> Query: Gets a query by its key.
        keys(): Returns the keys (query_ids) of all queries.
    """

    def __init__(
        self,
        experiment_name: str,
        save_path: str | None = None,
        evaluations_cache_path: str | None = None,
        cache_evaluations: bool = True,
        save_on_disk: bool = True,
        verbose: bool = False,
        rich_print: bool = True,
        clear_evaluations: bool = False,
        queries_csv_path: str | None = None,
        csv_query_id_col: str | None = None,
        csv_query_text_col: str = "query",
        documents_csv_path: str | None = None,
        csv_document_id_col: str | None = None,
        csv_document_text_col: str = "document",
        answers_csv_path: str | None = None,
        csv_agent_col: str = "agent",
        csv_answer_text_col: str = "answer",
    ):
        """
        Initialize the experiment object with optional data and set up paths and caches.

        Args:
            experiment_name (str): The name of your experiment
            save_path (Optional[str]): A JSON file path to persist the experiment on disk.
                If not set, results will not be saved locally. If it already exists,
                we will try to load the state from that file.
            evaluations_cache_path (Optional[str]): A JSON Lines file path to persist the evaluators result on disk,
                to avoid re-computing evaluations. If not set, will create one based on experiment_name.
            cache_evaluations (bool, defaults to True): Whether or not to cache the evaluations on evaluations_cache_path,
            save_on_disk (bool, defaults to True): Whether to save the experiment to disk using the experiment_name.
            verbose (bool, defaults to False): Whether to print information about the experiment.
            rich_print (bool, defaults to True): Whether to print information about the experiment using the rich library.
            clear_evaluations (bool, defaults to False): If set to True, will clear all existing evaluations and
                re-compute as needed.
            queries_csv_path (Optional[str]): An optional path to a CSV file with the queries to be used.
            csv_query_id_col (Optional[str]): The column name for query ID in the CSV files.
            csv_query_text_col (str): The column name for query text in the queries CSV file.
            documents_csv_path (Optional[str]): An optional path to a CSV file with the retrieved documents for
                the queries.
            csv_document_id_col (Optional[str]): The column name for document ID in the documents CSV file.
            csv_document_text_col (str): The column name for document text in the documents CSV file.
            answers_csv_path (Optional[str]): An optional path to a CSV file with the answers for the queries.
            csv_agent_col (str, defaults to "agent"): The column name for agent name in the documents and answers
                CSV files.
            csv_answer_text_col (str, defaults to "answer"): The column name for answer text in the answers CSV file.



        The initialization process involves setting up paths and loading queries from different sources:
        1. If `csv_path` is provided, queries are loaded from the specified CSV file.
        2. If `csv_path` is not provided but `save_path` exists and points to a valid file, queries are loaded from
            the cache file.
        3. If both `csv_path` and `save_path` are provided, the method ensures that the queries loaded from the CSV
            file match those in the cache file.
        The method prioritizes loading queries from the CSV file over the cache file. If neither source is available,
            it relies on the `queries` attribute if it has been set.
        """
        self.cache_evaluations = cache_evaluations
        self.evaluations_cache_path = evaluations_cache_path
        self.experiment_name = experiment_name
        self.queries: dict[str, Query] = {}
        self.elo_tournaments: list[EloTournamentResult] = []
        self.rich_print = rich_print
        self.save_on_disk = save_on_disk
        self.save_path = save_path
        if verbose:
            logger.setLevel(logging.INFO)
            if len(logger.handlers) == 0:
                logger.addHandler(CLILogHandler(use_rich=rich_print))
        elif len(logger.handlers) > 0:
            logger.setLevel(logging.WARNING)
            logger.handlers = []
            logger.addHandler(CLILogHandler(use_rich=rich_print))
        if self.save_on_disk and not self.save_path:
            self.save_path = f"ragelo_cache/{self.experiment_name}.json"
            if not os.path.exists("ragelo_cache"):
                logger.info(f"Creating a cache file for the experiment at ragelo_cache/{self.experiment_name}.json")
                Path("ragelo_cache").mkdir(parents=True, exist_ok=True)

        elif self.save_path is None and not self.save_on_disk:
            logger.info("No save path provided. Will not save the experiment to disk.")

        if self.cache_evaluations and not self.evaluations_cache_path:
            if self.save_path:
                self.evaluations_cache_path = self.save_path.replace(".json", "_results.jsonl")
            else:
                self.evaluations_cache_path = f"ragelo_cache/{self.experiment_name}_results.jsonl"
                if not os.path.exists("ragelo_cache"):
                    os.makedirs("ragelo_cache", exist_ok=True)
            if not os.path.isfile(self.evaluations_cache_path):
                logger.info(f"Creating a cache file for the experiment's evaluations at{self.evaluations_cache_path}")
                Path(self.evaluations_cache_path).touch()

        if self.save_path and os.path.isfile(self.save_path):
            self.queries = self._load_from_cache(self.save_path)
        else:
            self.queries = {}
        if queries_csv_path:
            self.add_queries_from_csv(queries_csv_path, csv_query_id_col, csv_query_text_col, exist_ok=True)
        if documents_csv_path:
            self.add_documents_from_csv(
                documents_csv_path,
                csv_document_id_col,
                csv_document_text_col,
                csv_query_id_col,
                csv_agent_col,
                exist_ok=True,
            )
        if answers_csv_path:
            self.add_agent_answers_from_csv(
                answers_csv_path, csv_agent_col, csv_answer_text_col, csv_query_id_col, exist_ok=True
            )

        if self.evaluations_cache_path and os.path.isfile(self.evaluations_cache_path):
            self._load_results_from_cache(self.evaluations_cache_path)
        if clear_evaluations:
            self.__clear_all_evaluations()
        self.save()

    def add_query(
        self,
        query: Query | str,
        query_id: str | None = None,
        metadata: dict | None = None,
        force: bool = False,
    ) -> str:
        """
        Adds a query to the collection of queries.
        Args:
            query (Query | str): The query to be added. It can be an instance of the Query class or a string.
            query_id (str | None, optional): The identifier for the query.
                If not provided, a default ID will be generated.
            metadata (dict | None, optional): Additional metadata for the query. Defaults to None.
            force (bool, optional): Whether to overwrite the query if it already exists. Defaults to False.
        """

        if isinstance(query, Query):
            if query_id is not None and query.qid != query_id:
                warnings.warn(f"Query ID mismatch. Using ID from query object: {query.qid}")
            query_id = query.qid
            query_obj = query
        elif query_id is None:
            query_id = f"query_{len(self.queries) + 1}"
            warnings.warn(f"Query ID not provided. Using default query ID: {query_id}.")
            query_obj = Query(qid=query_id, query=query, metadata=metadata)
        else:
            query_obj = Query(qid=query_id, query=query, metadata=metadata)
            query_id = query_obj.qid
        if query_id in self.queries and not force:
            logger.info(f'Query with ID "{query_id}" already exists. Use force=True to overwrite')
            return query_id
        if query_id in self.queries and force:
            logger.info(f'Query with ID "{query_id}" already exists, but force was set to True. Overwriting.')
        self.queries[query_id] = query_obj
        return query_id

    def add_retrieved_doc(
        self,
        doc: Document | str,
        query_id: str | None = None,
        doc_id: str | None = None,
        score: float | None = None,
        agent: str | None = None,
        force: bool = False,
        exist_ok: bool = False,
    ):
        """
        Adds a retrieved document to a query.
        Args:
            doc (Document | str): The document to be added. It can be an instance of the Document class or a string.
            query_id (Optional[str]): The query used to retrieve the document.
            doc_id (Optional[str]): The identifier for the document. If not provided, a default ID will be generated.
            score (Optional[float]): The score of the document. Defaults to None.
            agent (Optional[str]): The agent that retrieved the document. Defaults to None.
            force (bool): Whether to overwrite the document if it already exists. Defaults to False.
            exist_ok (bool): If True, will not raise an error if the document already exists.
                Defaults to False.
        """
        if isinstance(doc, Document):
            if doc.qid != query_id and query_id is not None:
                logger.warning(f"Query ID mismatch. Using ID from document object: {doc.qid}")
                query_id = doc.qid
            if doc.did != doc_id and doc_id is not None:
                logger.warning(f"Document ID mismatch. Using ID from document object: {doc.did}")
            doc_id = doc.did
            query_id = doc.qid
        elif doc_id is None:
            raise ValueError("Document was provided as string, but Document ID was not provided")
        if query_id is None:
            raise ValueError(f"ID of the query to which document {doc_id} was retrieved was not provided.")
        if isinstance(doc, str):
            doc = Document(qid=query_id, did=doc_id, text=doc)
            query_id = doc.qid
        if query_id not in self.queries:
            raise ValueError(f"Query {query_id} not found in queries")
        self.queries[query_id].add_retrieved_doc(doc, score, agent, force, exist_ok)

    def add_agent_answer(
        self,
        answer: AgentAnswer | str,
        agent: str | None = None,
        query_id: str | None = None,
        force: bool = False,
        exist_ok: bool = False,
    ):
        if isinstance(answer, AgentAnswer):
            if answer.qid != query_id and query_id is not None:
                logger.warning(f"Query ID mismatch. Using ID from answer object: {answer.qid}")
            query_id = answer.qid
            if answer.agent != agent and agent is not None:
                logger.warning(f"Agent mismatch. Using agent from answer object: {answer.agent}")
        elif query_id is None:
            raise ValueError("Query ID not provided")
        else:
            if agent is None:
                agent = f"agent_{len(self.queries[query_id].answers) + 1}"
                logger.warning(f"Agent not provided. Using default agent name: {agent}.")
            if query_id is None:
                raise ValueError("Query ID not provided")
            answer = AgentAnswer(qid=query_id, agent=agent, text=answer)
        query_id = answer.qid
        if query_id not in self.queries:
            raise ValueError(f"Query {query_id} not found in queries")
        self.queries[query_id].add_agent_answer(answer, force=force, exist_ok=exist_ok)

    @overload
    def add_evaluation(
        self,
        evaluation: RetrievalEvaluatorResult,
        should_save: bool = True,
        should_print: bool = True,
        force: bool = False,
        exist_ok: bool = False,
    ): ...

    @overload
    def add_evaluation(
        self,
        evaluation: AnswerEvaluatorResult,
        should_save: bool = True,
        should_print: bool = True,
        force: bool = False,
        exist_ok: bool = False,
    ): ...

    @overload
    def add_evaluation(
        self,
        evaluation: EloTournamentResult,
        should_save: bool = True,
        should_print: bool = True,
        force: bool = False,
        exist_ok: bool = False,
    ): ...

    def add_evaluation(
        self,
        evaluation: RetrievalEvaluatorResult | AnswerEvaluatorResult | EloTournamentResult,
        should_save: bool = True,
        should_print: bool = False,
        force: bool = False,
        exist_ok: bool = False,
    ):
        """
        Add an evaluation to the queries and optionally save the result.
        Args:
            evaluation (RetrievalEvaluatorResult | AnswerEvaluatorResult | EloTournamentResult): The evaluation result to be added.
            should_save (bool): Whether to save the result to disk. Defaults to True.
            should_print (bool): Whether to print the result. Defaults to True.
            force (bool): Whether to overwrite an existing evaluation. Defaults to False.
            exist_ok (bool): Whether to warn if an evaluation already exists. Defaults to False.
        """
        if isinstance(evaluation, EloTournamentResult):
            self.elo_tournaments.append(evaluation)
            added = True
        else:
            qid = evaluation.qid
            if qid not in self.queries:
                raise ValueError(f"Query {qid} not found in queries")
            added = self.queries[qid].add_evaluation(evaluation, force=force, exist_ok=exist_ok)
        if not added:
            return
        if should_save:
            self.save_results(evaluation)
        if not should_print:
            return
        if isinstance(evaluation, EloTournamentResult):
            self._print_elo_tournament_result(evaluation)
        else:
            self._print_response(evaluation)

    def _print_elo_tournament_result(self, result: EloTournamentResult):
        scores = sorted(result.scores.items(), key=lambda x: x[1], reverse=True)
        if self.rich_print:
            rich.print("-------[bold white] Agents Elo Ratings [/bold white]-------")
        else:
            print("------- Agents Elo Ratings -------")
        for agent, rating in scores:
            std_dev = result.std_dev.get(agent, 0)
            if self.rich_print:
                rich.print(f"[bold white]{agent:<15}[/bold white]: {rating:.1f}(±{std_dev:.1f})")
            else:
                print(f"{agent:<15}: {rating:.1f} (±{std_dev:.1f})")

    def _print_response(
        self,
        response: EvaluatorResult,
    ):
        answer: float | str | dict | int
        if isinstance(response.answer, dict):
            # Print the answer in a more readable format
            answer = json.dumps(response.answer, indent=4, ensure_ascii=False)
        elif isinstance(response.answer, BaseModel):
            answer = response.answer.model_dump_json(indent=4)
        else:
            answer = str(response.answer)
        response_dict = response.model_dump()
        agent_a = response_dict.get("agent_a")
        agent_b = response_dict.get("agent_b")
        agent = response_dict.get("agent")
        qid = response_dict.get("qid")
        did = response_dict.get("did")
        raw_answer = response_dict.get("raw_answer")

        if self.rich_print:
            rich.print(f"[bold blue]🔎 Query ID[/bold blue]: {qid}")
            did = response_dict.get("did")
            if did:
                rich.print(f"[bold blue]📜 Document ID[/bold blue]: {did}")
            if agent_a and agent_b:
                rich.print(f"[bold bright_cyan] {agent_a:<18} [/bold bright_cyan] 🆚  [bold red] {agent_b}[/bold red]")
            elif agent:
                rich.print(f"[bold bright_cyan]🕵️ Agent[/bold bright_cyan]: {agent}")
            rich.print(f"[bold blue]Parsed Answer[/bold blue]: {answer}")
            rich.print("")
            return
        print(f"Query ID: {qid}")
        if did:
            print(f"Document ID: {did}")
        if agent_a and agent_b:
            print(f"{agent_a} vs {agent_b}")
        elif agent:
            print(f"Agent: {agent}")
        print(f"Raw Answer: {raw_answer}")
        print(f"Parsed Answer: {answer}")
        print("")

    def evaluate_retrieval(
        self,
        relevance_key: str = "relevance",
        metrics: list[str] = ["Precision@10", "nDCG@10", "Judged@10"],
        relevance_threshold: int = 0,
    ) -> dict[str, dict[str, float]]:
        """
        Evaluate the retrieval performance of agents using specified metrics.
        Args:
            relevance_key (str): The key in the answer object that contains an integer with the relevance of the
                document.
            metrics (list[str]): A list of metrics to use. defaults to ["Precision@10", "nDCG@10", "Judged@10"].
            relevance_threshold (int): The threshold above which a document is considered relevant (default is 0).
        Returns:
            dict[str, dict[str, float]]: A dictionary where keys are agent names and values are dictionaries of scores.
        Raises:
            ImportError: If the `ir_measures` package is not installed.
            ValueError: If a specified metric is not found in the `ir_measures` registry.
        Notes:
        - The function uses the `ir_measures` package to calculate the metrics.
        - If the `rich` package is installed, the results are printed in a formatted table.
        - If the `rich` package is not installed, the results are printed using plain `print`.

        Example:
        >>> results = evaluate_retrieval(relevance_key="relevance", metrics=["Precision@10", "nDCG@10"], relevance_threshold=1)
        >>> print(results)
        {
            "agent1": {
                "Precision@10": 0.5,
                "nDCG@10": 0.6
            },
            "agent2": {
                "Precision@10": 0.4,
                "nDCG@10": 0.5
            }
        }
        """

        try:
            import ir_measures
            from ir_measures import parse_measure
        except ImportError:
            raise ImportError("ir_measures is not installed. Please install it with `pip install ir-measures`")
        qrels = self.get_qrels(
            relevance_key=relevance_key,
            relevance_threshold=relevance_threshold,
        )
        runs = self.get_runs()
        measures = []
        for metric in metrics:
            try:
                measure = parse_measure(metric)
            except NameError:
                valid_metrics = list(ir_measures.measures.registry.keys())
                raise ValueError(f"Metric {metric} not found. Valid metrics are: {valid_metrics}")
            measures.append(measure)
        metrics = [str(m) for m in measures]
        results = {}
        for agent, run in runs.items():
            # Transform the keys of the results back to strings
            results[agent] = {
                str(k): v
                for k, v in ir_measures.calc_aggregate(measures, qrels, run).items()  # type: ignore
            }

        key_metric = metrics[0]
        max_agent_len = max([len(agent) for agent in results.keys()]) + 3
        max_metric_len = max([len(metric) for metric in metrics])
        sorted_agents = sorted(results.items(), key=lambda x: x[1][key_metric], reverse=True)
        if self.rich_print:
            rich.print("---[bold cyan] Retrieval Scores [/bold cyan] ---")
            if relevance_threshold > 0:
                rich.print(f"[bold yellow]Relevance threshold: {relevance_threshold}[/bold yellow]")
            header = f"[bold magenta]{'Agent Name':<{max_agent_len}}"
            header += "\t".join([f"{m:<{max_metric_len}}" for m in metrics])
            header += "[/bold magenta]"
            rich.print(f"[bold cyan]{header}[/bold cyan]")
            for agent, scores in sorted_agents:
                row = f"[bold white]{agent:<{max_agent_len}}[/bold white]"
                row += "\t".join([f"{scores[metric]:<{max_metric_len},.4f}" for metric in metrics])
                rich.print(row)

        else:
            print(results)
        return results

    def get_qrels(
        self,
        relevance_key: str | None = "relevance",
        relevance_threshold: int = 0,
        output_path: str | None = None,
        output_format: Literal["trec", "json"] = "trec",
    ) -> dict[str, dict[str, int]]:
        """
        Retrieve the qrels (query relevance judgments) for the queries.
        Args:
            relevance_key (str | None): The key to use for determining relevance. Defaults to "relevance".
            relevance_threshold (int): The threshold value for relevance. Defaults to 0.
        Returns:
            dict[str, dict[str, int]]: A dictionary where each key is a query ID and the value is another dictionary
                                       mapping document IDs to their relevance scores.
        """

        qrels = {}
        for qid, query in self.queries.items():
            qrels[qid] = query.get_qrels(relevance_key, relevance_threshold)
        if output_path:
            with open(output_path, "w") as f:
                if output_format.lower() == "trec":
                    for qid, qrel in qrels.items():
                        for did, rel in qrel.items():
                            f.write(f"{qid} Q0 {did} {rel}\n")
                elif output_format.lower() == "json":
                    json.dump(qrels, f)
                else:
                    raise ValueError(
                        f"Invalid output format for QRELS: {output_format}Valid options are 'trec' and 'json'"
                    )
        return qrels

    def get_runs(
        self,
        agents: list[str] | None = None,
        output_path: str | None = None,
        output_format: str = "trec",
    ) -> dict[str, dict[str, dict[str, float]]]:
        """
        Retrieve runs for specified agents.
        This method aggregates runs from all queries and organizes them by agent.
        Args:
            agents (list[str] | None): A list of agent names to filter the runs.
                If None, runs for all agents are retrieved.
            output_path (str | None): The path to save the runs to disk. If None, the runs are not saved.
            output_format (str): The format to save the runs. Defaults to "trec". If TREC, output_path should be a
                directory, and each agent will have a separate file.
        Returns:
            dict[str, dict[str, dict[str, float]]]: A nested dictionary where the first level keys are agent names,
            the second level keys are run identifiers, and the third level keys are metric names with their
                corresponding float values.
        """

        runs_by_agent: dict[str, dict[str, dict[str, float]]] = defaultdict(dict)
        for query in self.queries.values():
            runs = query.get_runs(agents)
            for agent, run in runs.items():
                runs_by_agent[agent].update(run)
        if output_path:
            if output_format.lower() == "trec":
                if not os.path.isdir(output_path):
                    os.makedirs(output_path, exist_ok=True)
                for agent, run in runs_by_agent.items():
                    with open(f"{output_path}/{agent}.run", "w") as f:
                        for qid, docs in run.items():
                            sorted_scores = sorted(docs.items(), key=lambda x: x[1], reverse=True)
                            for idx, (did, score) in enumerate(sorted_scores):
                                f.write(f"{qid} Q0 {did} {idx + 1} {score} {agent}\n")
            elif output_format.lower() == "json":
                with open(output_path, "w") as f:
                    json.dump(runs_by_agent, f)

        return runs_by_agent

    def save(self, output_path: str | None = None):
        """
        Persist the current state of the model to disk if caching is enabled.
        This method checks if caching is enabled by evaluating the `save_cache` attribute.
        If caching is enabled, it attempts to write the model's current state to the file
        specified by `cache_path`. If `cache_path` is not set, a ValueError is raised.
        Raises:
            ValueError: If `cache_path` is None and caching is enabled.
        """
        if not self.save_on_disk:
            return
        if output_path is None:
            if self.save_path is None:
                raise ValueError("Cannot save experiment without a save path")
            output_path = self.save_path
        output_dict: dict[str, Any] = {}

        output_dict["queries"] = {qid: query.model_dump() for qid, query in self.queries.items()}
        output_dict["experiment_name"] = self.experiment_name
        output_dict["elo_tournaments"] = [tournament.model_dump() for tournament in self.elo_tournaments]

        with open(output_path, "w") as f:
            json.dump(output_dict, f, indent=4, ensure_ascii=False)

    def save_results(
        self,
        result: EvaluatorResult | EloTournamentResult,
    ):
        """
        Persist the evaluation result to a cache file.
        This method writes the given evaluation result to a cache file specified by
        `self.evaluations_cache_path`. The result is serialized to JSON format and appended
        to the file. The type of the result (either "answer" or "retrieval") is included
        in the serialized data.
        Args:
            result (EvaluatorResult | EloTournamentResult): The evaluation result
                to be persisted to the cache.
        Raises:
            ValueError: If `self.evaluations_cache_path` is not set or if the type of `result`
                is not recognized.
        """

        if not self.cache_evaluations:
            return
        if self.evaluations_cache_path is None:
            raise ValueError("Results cache path not set. Cannot dump result")
        with open(self.evaluations_cache_path, "a+") as f:
            if isinstance(result, AnswerEvaluatorResult):
                result_type = "answer"
            elif isinstance(result, RetrievalEvaluatorResult):
                result_type = "retrieval"
            elif isinstance(result, EloTournamentResult):
                result_type = "elo_tournament"
            else:
                raise ValueError(f"Cannot save evaluation of type {type(result)} to cache")
            f.write(json.dumps({result_type: result.model_dump()}, ensure_ascii=False) + "\n")

    def add_queries_from_csv(
        self,
        file_path: str,
        query_id_column: str | None = None,
        query_text_column: str = "query",
        exist_ok: bool = False,
    ):
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"CSV file with queries {file_path} not found")
        read_queries = set()
        try:
            query_id_column = self.__infer_query_id_column(file_path)
        except ValueError:
            query_id_column = None

        for idx, row in enumerate(csv.DictReader(open(file_path))):
            if query_id_column is None:
                query_id_column = f"query_{idx}"
            else:
                qid = row.get(query_id_column, f"query_{idx}")
            if qid in read_queries or qid in self.queries:
                if not exist_ok:
                    logger.warning(f"Query with ID {qid} already read. Skipping")
                continue
            query_text = row[query_text_column].strip()
            metadata = {k: v for k, v in row.items() if k not in [query_id_column, query_text_column]}
            self.queries[qid] = Query(qid=qid, query=query_text, metadata=metadata)
            read_queries.add(qid)
        if len(read_queries) > 0:
            logger.info(f"Loaded {len(read_queries)} queries from {file_path}")

    def add_documents_from_csv(
        self,
        file_path: str,
        document_id_col: str | None = "did",
        document_text_col: str = "document",
        query_id_col: str | None = "qid",
        agent_col: str | None = None,
        exist_ok: bool = False,
    ):
        """
        Loads a list of retrieved documents from a csv file. The csv file should have the following columns:
        query_id, document_id, document_text, [agent], [metadata columns]
        Args:
            file_path (str): Path to the CSV file with the retrieved documents.
            document_id_col (str): Name of the column with the document id. Defaults to 'did'.
            query_id_col (str): Name of the column with the query id to which the document was retrieved.
                Defaults to 'qid'.
            document_text_col (str): Name of the column with the document text. Defaults to 'document_text'.
            agent_col (str): Name of the column with the agent name. If None, will not add the agent name.
                Defaults to None.
            exist_ok (bool): If True, will not raise an error if the document already exists.
                Defaults to False.
        """
        documents_read = 0
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"CSV file {file_path} not found")
        query_id_col = self.__infer_query_id_column(file_path, query_id_col)
        document_id_col = self.__infer_document_id_column(file_path, document_id_col)
        warned_queries = set()

        for line in csv.DictReader(open(file_path)):
            qid = line[query_id_col].strip()
            did = line[document_id_col].strip()
            text = line[document_text_col].strip()
            agent = line.get(agent_col)
            metadata = {
                k: v for k, v in line.items() if k not in [query_id_col, document_id_col, document_text_col, agent_col]
            }
            if qid not in self.queries:
                if qid not in warned_queries:
                    warned_queries.add(qid)
                    logger.warning(f"Query {qid} found in {file_path} but not found in queries. Skipping")
                continue
            if (doc_obj := self.queries[qid].retrieved_docs.get(did)) is None:
                doc_obj = Document(qid=qid, did=did, text=text)
                doc_obj.add_metadata(metadata)
                documents_read += 1
            if agent is not None:
                doc_obj.add_retrieved_by(agent, exist_ok=exist_ok)
            self.add_retrieved_doc(doc_obj, exist_ok=exist_ok)
        if documents_read > 0:
            logger.info(f"Loaded {documents_read} new documents from {file_path}")

    def add_retrieved_docs_from_runfile(
        self,
        run_file_path: str,
        corpus: dict[str, Document] | dict[str, str],
        top_k: int = 1000,
    ):
        """
        Adds the retrieved documents from a run file to the queries.

        A run file has, traditionally, the following format:
            qid, Q0, docid, rank, score, run_name
        The run name will be considered as the agent that retrieved the documents.
            The second column is usually ignored.
        Args:
            run_file_path (str): Path to the run file.
            corpus (dict[str, Document] | dict[str, str]): A dictionary with the documents.
                If a dictionary with just docid and text is provided, we will use it as a corpus.
                Otherwise, we will assume it's a dictionary with docid as key and Document object as value.
            top_k (int): The maximum number of documents to add for each query.
                Assumes that the run file is ordered by rank.
        """
        documents_read = set()
        missing_docs = set()
        docs_per_query = {}
        warned_queries = set()

        if isinstance(list(corpus.values())[0], str):
            corpus = {k: str(v) for k, v in corpus.items()}
            # corpus = {k: Document(did=k, text=str(v)) for k, v in corpus.items()}

        if not os.path.isfile(run_file_path):
            raise FileNotFoundError(f"Run file {run_file_path} not found")
        for line in open(run_file_path):
            qid, _, did, _, score, agent = line.strip().split()
            if did not in corpus:
                missing_docs.add(did)
                continue
            doc = corpus[did]
            if isinstance(doc, str):
                doc = Document(qid=qid, did=did, text=doc)
            if qid not in self.queries:
                if qid not in warned_queries:
                    warned_queries.add(qid)
                    logger.warning(f"Query {qid} not found in queries. Skipping")
                continue
            if qid not in docs_per_query:
                docs_per_query[qid] = 0
            if docs_per_query[qid] < top_k:
                self.add_retrieved_doc(doc, qid, score=float(score), agent=agent)
                docs_per_query[qid] += 1
                documents_read.add(did)
        if len(missing_docs) > 0:
            logger.warning(f"Loaded {len(documents_read)} documents. {len(missing_docs)} missing docs")

    def add_agent_answers_from_csv(
        self,
        file_path: str,
        agent_col: str = "agent",
        answer_col: str = "answer",
        query_id_col: str | None = None,
        exist_ok: bool = False,
    ):
        """
        Adds agent answers from a CSV file to the queries.
        The CSV file should have the following columns:
        query_id, agent, answer, [metadata columns]
        Args:
            csv_path (str): Path to the CSV file with the agent answers.
            agent_col (str): Name of the column with the agent name. Defaults to 'agent'.
            answer_col (str): Name of the column with the answer text. Defaults to 'answer'.
            query_id_col (str): Name of the column with the query id.
                If None, will try to infer the query id column. Defaults to None.
            exist_ok (bool): If True, will not raise an error if the answer already exists.
                Defaults to False.
        """
        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"Answers file {file_path} not found")
        answers_read = 0
        warned_queries = set()
        query_id_col = self.__infer_query_id_column(file_path, query_id_col)
        for line in csv.DictReader(open(file_path)):
            qid = line[query_id_col].strip()
            agent = line[agent_col].strip()
            answer = line[answer_col].strip()
            metadata = {k: v for k, v in line.items() if k not in [query_id_col, agent_col, answer_col]}
            if qid not in self.queries:
                if qid not in warned_queries:
                    warned_queries.add(qid)
                    logger.warning(f"Query {qid} found in {file_path} but not found in queries. Skipping")
                continue
            if agent not in self.queries[qid].answers:
                answer_obj = AgentAnswer(qid=qid, agent=agent, text=answer)
                answer_obj.add_metadata(metadata)
                self.add_agent_answer(answer_obj, exist_ok=exist_ok)
                answers_read += 1
            elif not exist_ok:
                logger.info(f"Answer from agent {agent} already exists in query {qid}. Skipping")
        if answers_read > 0:
            logger.info(f"Loaded {answers_read} answers from {file_path}")

    def add_answers_from_multiple_csv(
        self,
        answers_files: list[str],
        query_id_col: str | None = None,
        query_text_col: str = "query",
        answer_text_col: str = "answer",
    ):
        """Loads queries and answers from a list of CSVs with queries and answers generated by agents.
        We assume that each csv was generated by a single agent, and name each agent according to the file name.
        Args:
            answers_files (list[str]): List of paths to the CSV files with the answers.
            query_id_col (str): The column name with the query ID.
                If None, will try to infer the query ID column. Defaults to None.
            query_text_col (str): The column name with the query text. Defaults to 'query'.
            answer_text_col (str): The column name with the answer text. Defaults to 'answer'.
        """
        for f in answers_files:
            if query_id_col is None:
                query_id_col = self.__infer_query_id_column(f)
            if query_id_col is None:
                query_id_col = query_text_col
            p = Path(f)
            agent_name = p.stem

            for line in csv.DictReader(open(p)):
                qid = line[query_id_col].strip()
                query = line[query_text_col].strip()
                answer = line[answer_text_col].strip()
                if qid not in self.queries:
                    self.add_query(query, qid)
                self.add_agent_answer(answer, agent_name, qid)

    def __clear_all_evaluations(self, should_save: bool = False):
        logger.warning(f"Clearing all evaluations for {len(self)} queries")
        doc_eval_count = 0
        game_eval_count = 0
        answer_eval_count = 0
        for query in self.queries.values():
            for doc in query.retrieved_docs.values():
                if doc.evaluation is not None:
                    doc_eval_count += 1
                doc.evaluation = None
            for game in query.pairwise_games:
                if game.evaluation is not None:
                    game_eval_count += 1
                game.evaluation = None
            for answer in query.answers.values():
                if answer.evaluation is not None:
                    answer_eval_count += 1
                answer.evaluation = None
        self.elo_tournaments = []
        if doc_eval_count > 0 or game_eval_count > 0 or answer_eval_count > 0:
            logger.info(
                f"Cleared {doc_eval_count} document evaluations, {game_eval_count} game evaluations, "
                f"and {answer_eval_count} answer evaluations"
            )

        if should_save:
            if self.evaluations_cache_path and os.path.isfile(self.evaluations_cache_path):
                with open(self.evaluations_cache_path, "w"):
                    pass

            self.save()

    def _load_from_cache(self, cache_path: str) -> dict[str, Query]:
        if not os.path.isfile(cache_path):
            raise FileNotFoundError(f"Cache file {cache_path} not found")
        logger.info(f"Loading existing experiment from {cache_path}")
        with open(cache_path) as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError:
                raise ValueError(f"Cache file {cache_path} is not a valid JSON file")
        read_experiment_name = data.get("experiment_name")
        if read_experiment_name and read_experiment_name != self.experiment_name:
            logger.warning(f"Experiment name mismatch. Expected {self.experiment_name}. Found {read_experiment_name}")
        queries = {k: Query(**v) for k, v in data.get("queries").items()}
        for q in queries.values():
            for doc in q.retrieved_docs.values():
                if doc.evaluation is not None and not isinstance(doc.evaluation, RetrievalEvaluatorResult):
                    doc_eval_dict = doc.evaluation.model_dump()
                    doc_eval_dict["did"] = doc.did
                    doc.evaluation = RetrievalEvaluatorResult(**doc_eval_dict)
            for answer in q.answers.values():
                if answer.evaluation is not None and not isinstance(answer.evaluation, AnswerEvaluatorResult):
                    answer_eval_dict = answer.evaluation.model_dump()
                    answer.evaluation = AnswerEvaluatorResult(**answer_eval_dict)
        if "elo_tournaments" in data:
            self.elo_tournaments = [EloTournamentResult(**t) for t in data["elo_tournaments"]]
        return queries

    def _load_results_from_cache(self, cache_path: str):
        assert os.path.isfile(cache_path)
        for line in open(cache_path):
            result = json.loads(line)
            result_type = list(result.keys())[0]
            if result_type == "answer":
                result = AnswerEvaluatorResult(**result["answer"])
                if result.qid not in self.queries:
                    continue
            elif result_type == "retrieval":
                result = RetrievalEvaluatorResult(**result["retrieval"])
                if result.qid not in self.queries:
                    continue
            elif result_type == "elo_tournament":
                result = EloTournamentResult(**result["elo_tournament"])
            self.add_evaluation(result, should_save=False, should_print=False, exist_ok=True)

    def __len__(self):
        return len(self.queries)

    def __iter__(self):
        return iter(self.queries.values())

    def __getitem__(self, key: str) -> Query:
        return self.queries[key]

    def keys(self):
        return self.queries.keys()

    @staticmethod
    def __infer_query_id_column(file_path: str, query_id_col: str | None = None) -> str:
        """Infer the column name with the query id from a CSV file."""
        if query_id_col is not None:
            return query_id_col
        possible_qid_columns = [
            "query_id",
            "qid",
            "question_id",
            "q_id",
            "queryid",
            "questionid",
        ]
        with open(file_path) as f:
            reader = csv.DictReader(f)
            columns = reader.fieldnames
            if columns is not None:
                for col in possible_qid_columns:
                    if col in columns:
                        return col
        raise ValueError(f"Could not identify Query ID column for CSV file {file_path}")

    @staticmethod
    def __infer_document_id_column(file_path: str, doc_id_col: str | None = None) -> str:
        """Infer the column name with the doc id from a CSV file."""
        if doc_id_col is not None:
            return doc_id_col
        possible_did_columns = [
            "document_id",
            "did",
            "d_id",
            "documentid",
            "passage_id",
            "pid",
            "p_id",
            "passageid",
        ]
        with open(file_path) as f:
            reader = csv.DictReader(f)
            columns = reader.fieldnames
            if columns is not None:
                for col in possible_did_columns:
                    if col in columns:
                        return col
        raise ValueError(f"Could not identify Document ID column for CSV file {file_path}")

    @classmethod
    def load(cls, experiment_name: str, path: str):
        return cls(experiment_name=experiment_name, save_path=path)



from __future__ import annotations

import json
from typing import Any, Type

from openai import AsyncOpenAI
from pydantic import BaseModel
from tenacity import retry, stop_after_attempt, wait_random_exponential

from ragelo.llm_providers.base_llm_provider import BaseLLMProvider, LLMProviderFactory
from ragelo.types.configurations import OllamaConfiguration
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.types import LLMProviderTypes


@LLMProviderFactory.register(LLMProviderTypes.OLLAMA)
class OllamaProvider(BaseLLMProvider):
    """A Wrapper over the Ollama client."""

    config: OllamaConfiguration

    def __init__(
        self,
        config: OllamaConfiguration,
    ):
        super().__init__(config)
        self.__ollama_client = self.__get_ollama_client(config)

    @retry(wait=wait_random_exponential(min=1, max=120), stop=stop_after_attempt(1))
    async def call_async(
        self,
        input: LLMInputPrompt,
        response_schema: Type[BaseModel] | dict[str, Any] | None = None,
    ) -> LLMResponseType:
        """Calls the Ollama Local API asynchronously.

        Args:
            user_prompt: The user prompt to send to the model.
            system_prompt: The system prompt to send to the model.
            answer_format: The format of the answer to return.
            response_schema: The response schema for structured output.
        """
        messages = []
        call_kwargs = {
            "model": self.config.model,
            "temperature": self.config.temperature,
            "max_tokens": self.config.max_tokens,
            "seed": self.config.seed,
        }

        if input.system_prompt and input.messages:
            raise ValueError(
                "If the input to the LLMProvider is a list of messages, you should not provide the system prompt as a unique parameter. Please combine both in a single input list"
            )
        if input.system_prompt:
            messages.append({"role": "system", "content": input.system_prompt})
        if input.user_message:
            messages.append({"role": "user", "content": input.user_message})
        if input.messages:
            messages = input.messages
        if not messages:
            raise ValueError("No input provided")

        call_kwargs["messages"] = messages  # type: ignore
        if isinstance(response_schema, type(BaseModel)):
            raise NotImplementedError("Structured answer format is not supported on Ollama.")
        if isinstance(response_schema, dict):
            schema = json.dumps(response_schema, indent=4)
            messages[-1]["content"] += (
                f"\n\nYour output should be a JSON string that STRICTLY adheres to the following schema:\n{schema}"
            )
            call_kwargs["response_format"] = {"type": "json_object"}  # type: ignore
        answers = await self.__ollama_client.chat.completions.create(**call_kwargs)  # type: ignore

        if not answers.choices or not answers.choices[0].message or not answers.choices[0].message.content:
            raise ValueError("Ollama did not return any completions.")

        if isinstance(response_schema, dict):
            return LLMResponseType(
                raw_answer=answers.choices[0].message.content,
                parsed_answer=json.loads(answers.choices[0].message.content),
            )
        return LLMResponseType(
            raw_answer=answers.choices[0].message.content,
            parsed_answer=answers.choices[0].message.content,
        )

    @staticmethod
    def __get_ollama_client(ollama_config: OllamaConfiguration) -> AsyncOpenAI:
        return AsyncOpenAI(
            base_url=ollama_config.api_base,
            api_key=ollama_config.api_key,
        )



from __future__ import annotations

from pydantic import BaseModel


class LLMProviderConfig(BaseModel):
    api_key: str | None
    temperature: float | None = 0.1
    max_tokens: int = 2048
    seed: int | None = 42


class OpenAIConfiguration(LLMProviderConfig):
    org: str | None = None
    api_type: str | None = None
    api_base: str | None = None
    api_version: str | None = None
    model: str = "gpt-4.1-mini"


class OllamaConfiguration(LLMProviderConfig):
    org: str | None = None
    api_base: str | None = "http://localhost:11434/v1/"
    api_version: str | None = None
    api_key: str | None = "NoKeyNeeded"
    model: str = "gemma2:27b"



from __future__ import annotations

from pydantic import Field

from ragelo.types.configurations.answer_evaluator_configs import (
    PairwiseDomainExpertEvaluatorConfig,
    PairwiseEvaluatorConfig,
)
from ragelo.types.configurations.base_configs import BaseConfig
from ragelo.types.configurations.retrieval_evaluator_configs import (
    DomainExpertEvaluatorConfig,
    RDNAMEvaluatorConfig,
    ReasonerEvaluatorConfig,
)


class BaseCLIConfig(BaseConfig):
    experiment_name: str = Field(
        default="experiment",
        description="The name of the experiment to run. This is also used as the local cache file name",
    )
    data_dir: str = Field(
        default="data",
        description="The directory where the data is stored.",
    )
    queries_csv_file: str = Field(
        default="queries.csv",
        description="The path to the queries CSV file. The file should contain at least the following columns: "
        "qid, query. Any additional columns will be considered as metadata.",
    )
    verbose: bool = Field(
        default=True,
        description="Whether or not to be verbose and print all intermediate steps.",
    )
    output_file: str | None = Field(
        default=None,
        description="The path to the output file where the results will be saved.",
    )
    save_results: bool = Field(
        default=True,
        description="Whether or not to save the results to disk.",
    )
    rich_print: bool = Field(
        default=True,
        description="Use rich to print colorful outputs.",
    )


class CLIEvaluatorConfig(BaseCLIConfig):
    documents_csv_file: str = Field(
        default="documents.csv",
        description=(
            "The path to the documents CSV file. The file should contain at least the following columns: "
            "qid, did, document. Any additional columns will be considered as metadata."
        ),
    )
    answers_csv_file: str = Field(
        default="answers.csv",
        description="The path to the answers CSV file. The file should contain at least the following columns: "
        "qid, agent, answer. Any additional columns will be considered as metadata. Ignored on Retrieval Evaluators.",
    )


class CLIDomainExpertEvaluatorConfig(CLIEvaluatorConfig, DomainExpertEvaluatorConfig):
    expert_in: str = " "


class CLIReasonerEvaluatorConfig(CLIEvaluatorConfig, ReasonerEvaluatorConfig):
    pass


class CLIRDNAMEvaluatorConfig(CLIEvaluatorConfig, RDNAMEvaluatorConfig):
    pass


class CLIPairwiseDomainExpertEvaluatorConfig(CLIEvaluatorConfig, PairwiseDomainExpertEvaluatorConfig):
    add_reasoning: bool = Field(
        default=False,
        description="If set to True, a reasoning retrieval evaluator will run, and the reasoning of the quality "
        "  of the retrieved results will be included in the prompt for the pairwise games.",
    )
    expert_in: str = " "


class CLIPairwiseEvaluatorConfig(CLIEvaluatorConfig, PairwiseEvaluatorConfig):
    add_reasoning: bool = Field(
        default=False,
        description="If set to True, a reasoning retrieval evaluator will run, and the reasoning of the quality "
        "  of the retrieved results will be included in the prompt for the pairwise games.",
    )


class CLIConfig(BaseCLIConfig):
    documents_csv_file: str = Field(
        default="documents.csv",
        description="The path to the documents CSV file. The file should contain at least the following columns: "
        "qid, did, document. Any additional columns will be considered as metadata.",
    )
    answers_csv_file: str = Field(
        default="answers.csv",
        description="The path to the answers CSV file. The file should contain at least the following columns: "
        "qid, did, answer. Any additional columns will be considered as metadata.",
    )

    k: int = Field(default=100, description="Number of pairwise games to generate")
    initial_score: int = Field(default=1000, description="The initial Elo score for each agent")
    elo_k: int = Field(default=32, description="The K factor for the Elo ranking algorithm")
    bidirectional: bool = Field(default=False, description="Wether or not to run each game in both directions")
    model: str = Field(default="gpt-4.1-mini", description="The model to use for the LLM")



import json
import os

import pytest

from ragelo import Experiment, get_agent_ranker, get_answer_evaluator, get_llm_provider, get_retrieval_evaluator
from ragelo.types.evaluables import AgentAnswer, Document
from ragelo.types.query import Query
from ragelo.types.results import AnswerEvaluatorResult, RetrievalEvaluatorResult


class TestExperiment:
    def test_experiment_initialization(self, experiment):
        assert len(experiment) == 2
        assert "0" in experiment.keys()
        assert "1" in experiment.keys()

        # Check queries were loaded correctly
        assert experiment["0"].query == "What is the capital of Brazil?"
        assert experiment["1"].query == "What is the capital of France?"

        # Check documents were loaded correctly
        assert len(experiment["0"].retrieved_docs) == 2
        assert experiment["0"].retrieved_docs["0"].text == "Brasília is the capital of Brazil."

        # Check answers were loaded correctly
        assert len(experiment["0"].answers) == 2
        assert "agent1" in experiment["0"].answers
        assert "agent2" in experiment["0"].answers

    def test_add_query(self, empty_experiment):
        """Test adding queries manually"""
        # Add query as string
        qid = empty_experiment.add_query("Test query 1", query_id="test1")
        assert qid in empty_experiment.queries
        assert empty_experiment[qid].query == "Test query 1"

        # Add query as Query object
        query = Query(qid="test2", query="Test query 2")
        qid2 = empty_experiment.add_query(query)
        assert qid2 == "test2"
        assert empty_experiment[qid2].query == "Test query 2"

        # Test force parameter
        empty_experiment.add_query("New query", qid, force=False)
        empty_experiment.add_query("Forced query", qid, force=True)
        assert empty_experiment[qid].query == "Forced query"

    def test_add_retrieved_doc(self, empty_experiment):
        """Test adding retrieved documents manually"""
        qid = empty_experiment.add_query("Test query", query_id="test1")

        # Add document as string
        empty_experiment.add_retrieved_doc("Test document", query_id=qid, doc_id="doc1", score=0.5, agent="agent1")
        assert "doc1" in empty_experiment[qid].retrieved_docs
        assert empty_experiment[qid].retrieved_docs["doc1"].text == "Test document"

        # Add document as Document object
        doc = Document(qid=qid, did="doc2", text="Test document 2")
        empty_experiment.add_retrieved_doc(doc, agent="agent2")
        assert "doc2" in empty_experiment[qid].retrieved_docs

        # Test adding to non-existent query
        with pytest.raises(ValueError):
            empty_experiment.add_retrieved_doc("Test", "invalid_qid", "doc3")

    def test_add_agent_answer(self, empty_experiment):
        """Test adding agent answers manually"""
        qid = empty_experiment.add_query("Test query", query_id="test1")

        # Add answer as string
        empty_experiment.add_agent_answer("Test answer", agent="agent1", query_id=qid)
        assert "agent1" in empty_experiment[qid].answers
        assert empty_experiment[qid].answers["agent1"].text == "Test answer"

        # Add answer as AgentAnswer object
        answer = AgentAnswer(qid=qid, agent="agent2", text="Test answer 2")
        empty_experiment.add_agent_answer(answer)
        assert "agent2" in empty_experiment[qid].answers

        # Test adding to non-existent query
        with pytest.raises(ValueError):
            empty_experiment.add_agent_answer("Test", "agent3", "invalid_qid")

    def test_save_and_load(self, tmp_path, base_experiment_config):
        """Test saving and loading experiment state"""
        # Save experiment
        save_path = tmp_path / "test_experiment.json"
        base_experiment_config["save_on_disk"] = True
        base_experiment_config["save_path"] = str(save_path)
        experiment = Experiment(**base_experiment_config)
        experiment.save()

        # Load experiment
        loaded_experiment = Experiment(
            experiment_name="test_experiment",
            save_path=str(save_path),
            cache_evaluations=False,
            save_on_disk=True,
        )

        # Verify contents
        assert len(loaded_experiment) == len(experiment)
        assert list(loaded_experiment.keys()) == list(experiment.keys())
        for qid in experiment.keys():
            assert loaded_experiment[qid].query == experiment[qid].query
            assert len(loaded_experiment[qid].retrieved_docs) == len(experiment[qid].retrieved_docs)
            assert len(loaded_experiment[qid].answers) == len(experiment[qid].answers)

    def test_get_qrels(self, tmp_path, experiment):
        """Test getting relevance judgments"""
        qrels = experiment.get_qrels()

        # Basic structure check
        assert len(qrels) == 2  # Two queries
        assert "0" in qrels
        assert "1" in qrels

        # Test saving qrels
        output_path = tmp_path / "test_qrels.txt"
        experiment.get_qrels(output_path=str(output_path), output_format="trec")
        assert os.path.exists(output_path)

    def test_add_retrieval_evaluation(self, experiment, retrieval_evaluation, caplog):
        """Test adding retrieval evaluation"""
        # Add evaluation
        experiment.add_evaluation(retrieval_evaluation, should_save=False)

        # Verify evaluation was added
        query = experiment["0"]
        doc = query.retrieved_docs["0"]
        assert doc.evaluation is not None
        assert doc.evaluation.answer == 1
        assert doc.evaluation.raw_answer == "Document is relevant. Score: 1.0"

        # Test adding duplicate evaluation without force
        added = experiment.add_evaluation(retrieval_evaluation, should_save=False)
        assert "Document 0 in query 0 already has an evaluation." in caplog.text
        assert not added

        # Test adding duplicate evaluation with force
        modified_eval = RetrievalEvaluatorResult(qid="0", did="0", raw_answer="Modified evaluation", answer=2)
        experiment.add_evaluation(modified_eval, should_save=False, force=True)
        assert doc.evaluation.answer == 2

        # Test adding evaluation for non-existent query
        invalid_eval = RetrievalEvaluatorResult(qid="999", did="0", raw_answer="Invalid", answer=1)
        with pytest.raises(ValueError):
            experiment.add_evaluation(invalid_eval)

    def test_add_answer_evaluation(self, experiment, answer_evaluation, caplog):
        """Test adding answer evaluation"""
        # Add evaluation
        experiment.add_evaluation(answer_evaluation, should_save=False)

        # Verify evaluation was added
        query = experiment["0"]
        answer = query.answers["agent1"]
        assert answer.evaluation is not None
        assert answer.evaluation.answer == {"quality": 1.0, "relevance": 0.8}
        assert answer.evaluation.raw_answer == "Answer is good. Scores: {'quality': 1.0, 'relevance': 0.8}"
        assert not answer.evaluation.pairwise

        # Test adding duplicate evaluation without force
        added = experiment.add_evaluation(answer_evaluation, should_save=False)
        assert not added
        assert "Agent agent1 in query 0 already has an evaluation." in caplog.text

        # Test adding duplicate evaluation with force
        modified_eval = AnswerEvaluatorResult(
            qid="0",
            agent="agent1",
            pairwise=False,
            raw_answer="Modified evaluation",
            answer={"quality": 2.0, "relevance": 1.0},
        )
        experiment.add_evaluation(modified_eval, should_save=False, force=True)
        assert answer.evaluation.answer == {"quality": 2.0, "relevance": 1.0}

    def test_add_pairwise_evaluation(self, experiment, pairwise_answer_evaluation):
        """Test adding pairwise answer evaluation"""
        # Add evaluation
        experiment.add_evaluation(pairwise_answer_evaluation, should_save=False)

        # Verify evaluation was added
        query = experiment["0"]
        assert len(query.pairwise_games) == 1
        game = query.pairwise_games[0]
        assert game.evaluation is not None
        assert game.evaluation.answer == "A"
        assert game.evaluation.raw_answer == "Answer [[A]] is better than [[B]]"
        assert game.evaluation.pairwise
        assert game.evaluation.agent_a == "agent1"
        assert game.evaluation.agent_b == "agent2"

    def test_add_elo_tournament(self, experiment, elo_tournament_result):
        """Test adding Elo tournament results"""
        # Add evaluation
        experiment.add_evaluation(elo_tournament_result, should_save=False)

        # Verify tournament was added
        assert len(experiment.elo_tournaments) == 1
        tournament = experiment.elo_tournaments[0]

        # Check tournament results
        assert tournament.agents == ["agent1", "agent2"]
        assert tournament.scores == {"agent1": 1200, "agent2": 1000}
        assert tournament.games_played == {"agent1": 1, "agent2": 1}
        assert tournament.wins == {"agent1": 1, "agent2": 0}
        assert tournament.loses == {"agent1": 0, "agent2": 1}
        assert tournament.ties == {"agent1": 0, "agent2": 0}
        assert tournament.total_games == 2
        assert tournament.total_tournaments == 1

    def test_save_evaluations(
        self,
        tmp_path,
        base_experiment_config,
        retrieval_evaluation,
        answer_evaluation,
        pairwise_answer_evaluation,
        elo_tournament_result,
    ):
        """Test saving evaluations to disk"""
        # Set up save paths
        results_path = tmp_path / "test_results.jsonl"
        base_experiment_config["evaluations_cache_path"] = str(results_path)
        base_experiment_config["cache_evaluations"] = True
        experiment = Experiment(**base_experiment_config)

        # Add and save evaluations
        experiment.add_evaluation(retrieval_evaluation, should_save=True)
        experiment.add_evaluation(answer_evaluation, should_save=True)
        experiment.add_evaluation(pairwise_answer_evaluation, should_save=True)
        experiment.add_evaluation(elo_tournament_result, should_save=True)

        # Verify results were saved
        assert results_path.exists()

        # Load and verify saved results
        with open(results_path) as f:
            lines = f.readlines()
            assert len(lines) == 4

            # Parse each line and verify content types
            results = [json.loads(line) for line in lines]
            result_types = [list(r.keys())[0] for r in results]
            assert "retrieval" in result_types
            assert "answer" in result_types
            assert "elo_tournament" in result_types

    def test_clear_evaluations(
        self,
        experiment,
        retrieval_evaluation,
        answer_evaluation,
        pairwise_answer_evaluation,
        elo_tournament_result,
    ):
        """Test clearing all evaluations"""
        # Add evaluations
        experiment.add_evaluation(retrieval_evaluation, should_save=False)
        experiment.add_evaluation(answer_evaluation, should_save=False)
        experiment.add_evaluation(pairwise_answer_evaluation, should_save=False)
        experiment.add_evaluation(elo_tournament_result, should_save=False)

        # Clear evaluations
        experiment._Experiment__clear_all_evaluations()

        # Verify evaluations were cleared
        query = experiment["0"]
        assert query.retrieved_docs["0"].evaluation is None
        assert query.answers["agent1"].evaluation is None
        assert len(experiment.elo_tournaments) == 0

    def test_get_runs(self, tmp_path, experiment_with_retrieval_scores):
        """Test getting retrieval runs"""
        runs = experiment_with_retrieval_scores.get_runs()

        # Basic structure check
        assert len(runs) > 0  # At least one agent

        # Test saving runs
        output_dir = tmp_path / "test_runs"
        experiment_with_retrieval_scores.get_runs(output_path=str(output_dir), output_format="trec")
        assert os.path.exists(output_dir)
        import shutil

        shutil.rmtree(output_dir)  # Cleanup

    def test_iteration(self, experiment):
        """Test iteration over queries"""
        queries = list(experiment)
        assert len(queries) == 2
        assert all(isinstance(q, Query) for q in queries)

    def test_add_queries_from_csv(self, empty_experiment):
        """Test adding queries from CSV"""
        empty_experiment.add_queries_from_csv("tests/data/queries.csv")
        assert len(empty_experiment) == 2
        assert "0" in empty_experiment.keys()
        assert "1" in empty_experiment.keys()

    def test_add_documents_from_csv(self, empty_experiment):
        """Test adding documents from CSV"""
        empty_experiment.add_queries_from_csv("tests/data/queries.csv")
        empty_experiment.add_documents_from_csv("tests/data/documents.csv")

        assert len(empty_experiment["0"].retrieved_docs) == 2
        assert len(empty_experiment["1"].retrieved_docs) == 2

    def test_add_agent_answers_from_csv(self, empty_experiment):
        """Test adding answers from CSV"""
        empty_experiment.add_queries_from_csv("tests/data/queries.csv")
        empty_experiment.add_agent_answers_from_csv("tests/data/answers.csv")

        assert len(empty_experiment["0"].answers) == 2
        assert len(empty_experiment["1"].answers) == 2

    @pytest.mark.requires_openai
    def test_readme_example(self, capsys):
        """Test the README example"""
        if os.path.exists("ragelo_cache/A_really_cool_RAGElo_experiment.json"):
            os.remove("ragelo_cache/A_really_cool_RAGElo_experiment.json")
        if os.path.exists("ragelo_cache/A_really_cool_RAGElo_experiment_results.jsonl"):
            os.remove("ragelo_cache/A_really_cool_RAGElo_experiment_results.jsonl")

        experiment = Experiment(experiment_name="A_really_cool_RAGElo_experiment")
        # Add two user queries. Alternatively, we can load them from a csv file with .add_queries_from_csv()
        experiment.add_query("What is the capital of Brazil?", query_id=0)
        experiment.add_query("What is the capital of France?", query_id=1)

        # Add four documents retrieved for these queries. Alternatively, we can load them from a csv file with .add_documents_from_csv()
        experiment.add_retrieved_doc("Brasília is the capital of Brazil", query_id=0, doc_id=0)
        experiment.add_retrieved_doc("Rio de Janeiro used to be the capital of Brazil.", query_id=0, doc_id=1)
        experiment.add_retrieved_doc("Paris is the capital of France.", query_id=1, doc_id=2)
        experiment.add_retrieved_doc("Lyon is the second largest city in France.", query_id=1, doc_id=3)

        # Add the answers generated by agents
        experiment.add_agent_answer("Brasília is the capital of Brazil, according to [0].", agent="agent1", query_id=0)
        experiment.add_agent_answer(
            "According to [1], Rio de Janeiro used to be the capital of Brazil, until the 60s.",
            agent="agent2",
            query_id=0,
        )
        experiment.add_agent_answer("Paris is the capital of France, according to [2].", agent="agent1", query_id=1)
        experiment.add_agent_answer(
            "According to [3], Lyon is the second largest city in France. Meanwhile, Paris is its capital [2].",
            agent="agent2",
            query_id=1,
        )

        llm_provider = get_llm_provider("openai", model="gpt-4.1-nano")

        retrieval_evaluator = get_retrieval_evaluator("reasoner", llm_provider=llm_provider, rich_print=True)
        answer_evaluator = get_answer_evaluator("pairwise", llm_provider=llm_provider, rich_print=True)

        elo_ranker = get_agent_ranker("elo", verbose=True)

        # Evaluate the retrieval results.
        retrieval_evaluator.evaluate_experiment(experiment)

        # With the retrieved documents evaluated, evaluate the quality of the answers. using the pairwise evaluator
        answer_evaluator.evaluate_experiment(experiment)
        elo_ranker.run(experiment)
        captured = capsys.readouterr()
        assert "Agents Elo Ratings" in captured.out
        assert "agent1" in captured.out
        assert "agent2" in captured.out
        assert os.path.exists("ragelo_cache/A_really_cool_RAGElo_experiment.json")
        assert os.path.exists("ragelo_cache/A_really_cool_RAGElo_experiment_results.jsonl")
        os.remove("ragelo_cache/A_really_cool_RAGElo_experiment.json")
        os.remove("ragelo_cache/A_really_cool_RAGElo_experiment_results.jsonl")



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field


class PairWiseAnswerAnswerFormat(BaseModel):
    answer_a_analysis: str = Field(..., description="A string with your analysis of assistant A's answer")
    answer_b_analysis: str = Field(..., description="A string with your analysis of assistant B's answer")
    comparison_reasoning: str = Field(
        ..., description="A string with your comparison between the two answers and their differences"
    )
    winner: Literal["A", "B", "C"] = Field(..., description="The winner of the pairwise comparison.")


class RetrievalAnswerEvaluatorFormat(BaseModel):
    reasoning: str = Field(..., description="A concise explanation and reasoning of the relevance of the document.")
    score: int = Field(
        ...,
        description="Your relevance score for the document. 0 for non-relevant, 1 for somewhat relevant and 2 for highly relevant.",
    )


class RDNAMAnswerEvaluatorFormat(BaseModel):
    overall: float = Field(
        ...,
        description="An number between 0 and 2 representing the score of the document.",
    )
    intent_match: float | None = Field(
        ...,
        description="An number between 0 and 2 representing the match of the document to the query intent.",
    )
    trustworthiness: float | None = Field(
        ..., description="An number between 0 and 2 representing the trustworthiness of the document."
    )


class RDNAMAnswerNoAspects(BaseModel):
    overall: float = Field(
        ...,
        description="An number between 0 and 2 representing the score of the document.",
    )


class RDNAMMultipleAnnotatorsAnswer(BaseModel):
    annotator_1: RDNAMAnswerEvaluatorFormat
    annotator_2: RDNAMAnswerEvaluatorFormat
    annotator_3: RDNAMAnswerEvaluatorFormat
    annotator_4: RDNAMAnswerEvaluatorFormat
    annotator_5: RDNAMAnswerEvaluatorFormat


class RDNAMMultipleAnnotatorsAnswerNoAspects(BaseModel):
    annotator_1: RDNAMAnswerNoAspects
    annotator_2: RDNAMAnswerNoAspects
    annotator_3: RDNAMAnswerNoAspects
    annotator_4: RDNAMAnswerNoAspects
    annotator_5: RDNAMAnswerNoAspects



"""Answer Evaluator for conversations between two agents."""

from ragelo.evaluators.answer_evaluators.base_answer_evaluator import AnswerEvaluatorFactory
from ragelo.evaluators.answer_evaluators.pairwise_evaluator import PairwiseAnswerEvaluator
from ragelo.types.configurations import PairwiseEvaluatorConfig
from ragelo.types.evaluables import PairwiseGame
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.types import AnswerEvaluatorTypes
from ragelo.utils import string_to_template


@AnswerEvaluatorFactory.register(AnswerEvaluatorTypes.CHAT_PAIRWISE)
class ChatPairwiseEvaluator(PairwiseAnswerEvaluator):
    config: PairwiseEvaluatorConfig
    system_prompt = string_to_template("""
        Please act as an impartial judge and evaluate the quality of the responses provided by two AI assistants tasked to answer the question of a user, based on a set of documents retrieved by a search engine that may or may not be relevant to the question.
        When available, answers will cite specific documents by placing their IDs into square brackets.
        {%- if doc and annotation %}
        You will be provided with the text of each reference document and its relevance evaluation.
        {%- elif doc %}
        You will be provided with the text of each reference document.
        {%- elif annotation %}
        For each cited document, you will be provided with its relevance evaluation
        {%- endif %}
        {%- if not reasoning %}
        Each document is scored in a scale of 0 to 2, where:
            - 0: the document is not relevant to the query
            - 1: the document is somewhat relevant to the query
            - 2: the document is highly relevant to the query
        {%- endif %}
        You should choose the assistant that best answers the user's question.

        ## Evaluation Guidelines
        Your evaluation should consider the evaluation objectives listed below:
        - If an answer contains non-relevant details, it should not be preferred over one that only use relevant information.
        - Avoid any position biases and ensure that the order in which the responses were presented does not 
        - Do not allow the length of the responses to influence your evaluation.
        - Be as objective as possible.

        ## Workflow
        - Begin your evaluation by examining each agent separately and explaining if each answer is useful towards satisfying the user's intent at each iteration.
        - Then, provide a short explanation how the agent performed overall based on the evaluation objectives
        - Finally, compare the conversations of the two agents and provide a short explanation on their differences.
        - After providing your explanation, output your final verdict by strictly following this format: 'A' if assistant A is better, 'B' if assistant B is better, and 'C' for a tie.
        """)
    user_prompt = string_to_template("""
        [User Intent]
        {{ query.query }}

        [Evaluation Objectives]
        {{ factors }}

        {%- if documents %}
        [Reference Documents]
        {%- for d in documents %}
        {%- if doc and (annotation or reasoning) %}
            Document ID: [{{ d.did }}]
            Content: {{ d.text }}
            Relevance: {% if reasoning %} {{ d.evaluation.answer.reasoning }} {% else %} {{ d.evaluation.answer.score }} {% endif %}
        ------------------
        {%- elif doc %}
            [{{ d.did }}]: {{ d.text }}
        {%- elif annotation %}
            [{{d.did }}] {% if reasoning %} {{ d.evaluation.answer.reasoning }} {% else %} {{ d.evaluation.answer.score }} {% endif %}"
        {% endif -%}
        {% endfor %}
        {% endif -%}

        [The Start of Conversation with Assistant A]
        {% for msg in game.agent_a_answer.conversation -%}
        {{ msg }}
        {% endfor %}
        [The End of Conversation with Assistant A]

        [The Start of Conversation with Assistant B]
        {% for msg in game.agent_b_answer.conversation -%}
        {{ msg }}
        {% endfor %}
        [The End of Conversation with Assistant B]""")

    def _build_message_pairwise(self, query: Query, game: PairwiseGame) -> LLMInputPrompt:
        documents = self._filter_documents(query)

        if not game.agent_a_answer.conversation or not game.agent_b_answer.conversation:
            raise ValueError("The conversation of the agents cannot be empty for the chat_pairwise evaluator")

        context = {
            "factors": self.config.factors,
            "query": query,
            "documents": documents,
            "game": game,
            "doc": self.config.include_raw_documents,
            "annotation": self.config.include_relevance_score,
            "reasoning": self.config.include_relevance_reasoning,
        }
        return LLMInputPrompt(
            system_message=self.system_prompt.render(**context),
            user_message=self.user_prompt.render(**context),
        )



from __future__ import annotations

from typing import Any, Type
from unittest.mock import AsyncMock

import pytest
from openai import AsyncOpenAI
from pydantic import BaseModel

from ragelo.llm_providers.base_llm_provider import BaseLLMProvider
from ragelo.llm_providers.openai_client import OpenAIConfiguration
from ragelo.types.answer_formats import (
    PairWiseAnswerAnswerFormat,
    RDNAMAnswerEvaluatorFormat,
    RDNAMMultipleAnnotatorsAnswer,
    RetrievalAnswerEvaluatorFormat,
)
from ragelo.types.configurations import (
    BaseAnswerEvaluatorConfig,
    BaseEvaluatorConfig,
    BaseRetrievalEvaluatorConfig,
    CustomPromptAnswerEvaluatorConfig,
    CustomPromptEvaluatorConfig,
    DomainExpertEvaluatorConfig,
    FewShotEvaluatorConfig,
    LLMProviderConfig,
    PairwiseDomainExpertEvaluatorConfig,
    PairwiseEvaluatorConfig,
    RDNAMEvaluatorConfig,
)
from ragelo.types.configurations.retrieval_evaluator_configs import FewShotExample
from ragelo.types.evaluables import ChatMessage
from ragelo.types.experiment import Experiment
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.results import AnswerEvaluatorResult, EloTournamentResult, RetrievalEvaluatorResult
from ragelo.types.types import AnswerEvaluatorTypes, RetrievalEvaluatorTypes
from ragelo.utils import string_to_template


def pytest_addoption(parser):
    parser.addoption(
        "--runopenai",
        action="store_true",
        default=False,
        help="run tests that requires an OpenAI API key",
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runopenai"):
        # --runopenai given in cli: do not skip tests that will call OpenAI
        return
    skip_openai = pytest.mark.skip(reason="need --runopenai option to run")
    for item in items:
        if "requires_openai" in item.keywords:
            item.add_marker(skip_openai)


@pytest.fixture
def openai_client_config():
    return OpenAIConfiguration(
        api_key="fake key",
        org="fake org",
        api_type="open_ai",
        api_base=None,
        api_version=None,
        model="fake model",
    )


@pytest.fixture
def llm_provider_config():
    return LLMProviderConfig(
        api_key="fake key",
    )


def answer_model_factory(input: LLMInputPrompt, response_schema, **kwargs):
    # Mirror OpenAIProvider behavior: response format is inferred from response_schema
    if response_schema == RetrievalAnswerEvaluatorFormat:
        return LLMResponseType(
            raw_answer='{"reasoning": "The document is very relevant", "score": 2}',
            parsed_answer=response_schema(
                reasoning="The document is very relevant",
                score=2,
            ),
        )
    if response_schema == PairWiseAnswerAnswerFormat:
        return LLMResponseType(
            raw_answer='{"answer_a_analysis": "The document is very relevant", "answer_b_analysis": "The document is not relevant", "comparison_reasoning": "The document is more relevant for the user question", "winner": "A"}',
            parsed_answer=response_schema(
                answer_a_analysis="Answer A is good",
                answer_b_analysis="Answer B is bad",
                comparison_reasoning="Answer A is better than Answer B",
                winner="A",
            ),
        )
    if response_schema == AnswerFormat:
        return LLMResponseType(
            raw_answer='{"keyA": "valueA", "keyB": "valueB"}',
            parsed_answer=response_schema(keyA="valueA", keyB="valueB"),
        )
    if response_schema == AnswerEvaluatorFormat:
        return LLMResponseType(
            raw_answer='{"quality": 1, "trustworthiness": 0, "originality": 0}',
            parsed_answer=response_schema(quality=1, trustworthiness=0, originality=0),
        )
    elif isinstance(response_schema, dict):
        return LLMResponseType(
            raw_answer='{"score": 1.0}',
            parsed_answer={"score": 1.0},
        )
    else:
        return LLMResponseType(
            raw_answer='{"keyA": "valueA", "keyB": "valueB"}',
            parsed_answer='{"keyA": "valueA", "keyB": "valueB"}',
        )


class MockLLMProvider(BaseLLMProvider):
    def __init__(self, config):
        super().__init__(config)
        self.async_call_mocker = AsyncMock()

    @classmethod
    def from_config(cls, config: LLMProviderConfig):
        return cls(config)

    async def call_async(
        self,
        input: LLMInputPrompt,
        response_schema: Type[BaseModel] | dict[str, Any] | None = None,
    ) -> LLMResponseType:
        # Record the call for assertions in tests
        if self.async_call_mocker.side_effect is not None:
            # If a custom side effect is set by a test/fixture, use it
            return await self.async_call_mocker(input, response_schema)
        else:
            # Still record the call (without relying on its return value)
            try:
                await self.async_call_mocker(input, response_schema)
            except Exception:
                pass
            return answer_model_factory(input, response_schema)


@pytest.fixture
def responses_api_mock(mocker, answer_format):
    """Factory helpers to mock the OpenAI Responses API.

    The OpenAIProvider chooses between responses.create (TEXT/JSON) and
    responses.parse (STRUCTURED) based on the provided response_schema.
    This fixture exposes helpers so the client mock can return suitable
    objects depending on call kwargs.
    """
    holder = mocker.Mock()

    def create_text_response():
        resp = mocker.Mock()
        resp.output_text = "fake response"
        return resp

    def create_json_response():
        resp = mocker.Mock()
        resp.output_text = '{"keyA": "valueA", "keyB": "valueB"}'
        return resp

    def parse_structured_response():
        resp = mocker.Mock()
        resp.output_text = '{"keyA": "valueA", "keyB": "valueB"}'
        resp.output_parsed = answer_format(keyA="valueA", keyB="valueB")
        return resp

    holder.create_text_response = create_text_response
    holder.create_json_response = create_json_response
    holder.parse_structured_response = parse_structured_response

    return holder


@pytest.fixture
def openai_client_mock(mocker, responses_api_mock):
    openai_client = mocker.AsyncMock(AsyncOpenAI)

    # Mock the responses API
    type(openai_client).responses = mocker.AsyncMock()

    # responses.create returns TEXT by default; if JSON schema is requested, return JSON
    def create_side_effect(*args, **kwargs):
        text = kwargs.get("text")
        text_fmt = kwargs.get("text_format")
        if isinstance(text, dict):
            return responses_api_mock.create_json_response()
        elif isinstance(text_fmt, type(BaseModel)):
            return responses_api_mock.parse_structured_response()
        return responses_api_mock.create_text_response()

    openai_client.responses.create = mocker.AsyncMock(side_effect=create_side_effect)
    openai_client.responses.parse = mocker.AsyncMock(side_effect=create_side_effect)

    return openai_client


@pytest.fixture
def base_experiment_config():
    config = {
        "experiment_name": "test_experiment",
        "save_on_disk": False,
        "cache_evaluations": False,
        "queries_csv_path": "tests/data/queries.csv",
        "documents_csv_path": "tests/data/documents.csv",
        "answers_csv_path": "tests/data/answers.csv",
        "rich_print": True,
        "verbose": True,
    }
    return config


@pytest.fixture
def experiment(base_experiment_config):
    return Experiment(**base_experiment_config)


@pytest.fixture
def experiment_with_conversations_and_reasonings(experiment):
    experiment.queries["0"].answers["agent1"].conversation = [
        ChatMessage(sender="user", content="What is the capital of Brazil?"),
        ChatMessage(
            sender="agent1",
            content="Brasília is the capital of Brazil, according to [0].",
        ),
    ]
    experiment.queries["0"].answers["agent2"].conversation = [
        ChatMessage(sender="user", content="What is the capital of Brazil?"),
        ChatMessage(
            sender="agent2",
            content="According to [1], Rio de Janeiro used to be the capital of Brazil, until the 60s.",
        ),
    ]
    experiment.queries["1"].answers["agent1"].conversation = [
        ChatMessage(sender="user", content="What is the capital of France?"),
        ChatMessage(sender="agent1", content="Paris is the capital of France, according to [2]."),
    ]
    experiment.queries["1"].answers["agent2"].conversation = [
        ChatMessage(sender="user", content="What is the capital of France?"),
        ChatMessage(
            sender="agent2",
            content="According to [3], Lyon is the second largest city in France. Meanwhile, Paris is its capital [2].",
        ),
    ]
    experiment.queries["0"].retrieved_docs["0"].evaluation = RetrievalEvaluatorResult(
        qid=experiment.queries["0"].qid,
        did="0",
        raw_answer="The document is very relevant as it directly answers the user's question about the capital of Brazil",
        answer="The document is very relevant as it directly answers the user's question about the capital of Brazil",
    )
    experiment.queries["0"].retrieved_docs["1"].evaluation = RetrievalEvaluatorResult(
        qid=experiment.queries["0"].qid,
        did="1",
        raw_answer="The document is somewhat relevant as it provides historical information about the capital of Brazil, but it does not provide the current capital.",
        answer="The document is somewhat relevant as it provides historical information about the capital of Brazil, but it does not provide the current capital.",
    )
    experiment.queries["1"].retrieved_docs["2"].evaluation = RetrievalEvaluatorResult(
        qid=experiment.queries["1"].qid,
        did="2",
        raw_answer="The document is very relevant as it directly answers the user's question about the capital of France.",
        answer="The document is very relevant as it directly answers the user's question about the capital of France.",
    )
    experiment.queries["1"].retrieved_docs["3"].evaluation = RetrievalEvaluatorResult(
        qid=experiment.queries["1"].qid,
        did="3",
        raw_answer="The document is not relevant to the user question as it does not provide information about the capital of France.",
        answer="The document is not relevant to the user question as it does not provide information about the capital of France.",
    )
    return experiment


@pytest.fixture
def experiment_with_retrieval_scores(experiment):
    experiment.queries["0"].retrieved_docs["0"].retrieved_by["agent1"] = 1.0
    experiment.queries["0"].retrieved_docs["0"].retrieved_by["agent2"] = 0.5
    experiment.queries["0"].retrieved_docs["1"].retrieved_by["agent1"] = 0.5
    experiment.queries["0"].retrieved_docs["1"].retrieved_by["agent2"] = 1.0
    experiment.queries["1"].retrieved_docs["2"].retrieved_by["agent1"] = 1.0
    experiment.queries["1"].retrieved_docs["2"].retrieved_by["agent2"] = 0.5
    experiment.queries["1"].retrieved_docs["3"].retrieved_by["agent1"] = 0.5
    experiment.queries["1"].retrieved_docs["3"].retrieved_by["agent2"] = 1.0
    return experiment


@pytest.fixture
def empty_experiment(base_experiment_config):
    base_config = base_experiment_config.copy()
    base_config.pop("queries_csv_path")
    base_config.pop("documents_csv_path")
    base_config.pop("answers_csv_path")
    return Experiment(**base_config)


@pytest.fixture
def retrieval_evaluation():
    return RetrievalEvaluatorResult(
        qid="0",
        did="0",
        raw_answer="Document is relevant. Score: 1.0",
        answer=1,
    )


@pytest.fixture
def answer_evaluation():
    return AnswerEvaluatorResult(
        qid="0",
        agent="agent1",
        pairwise=False,
        raw_answer="Answer is good. Scores: {'quality': 1.0, 'relevance': 0.8}",
        answer={"quality": 1.0, "relevance": 0.8},
    )


@pytest.fixture
def pairwise_answer_evaluation():
    return AnswerEvaluatorResult(
        qid="0",
        agent_a="agent1",
        agent_b="agent2",
        pairwise=True,
        raw_answer="Answer [[A]] is better than [[B]]",
        answer="A",
    )


@pytest.fixture
def elo_tournament_result():
    return EloTournamentResult(
        agents=["agent1", "agent2"],
        scores={"agent1": 1200, "agent2": 1000},
        games_played={"agent1": 1, "agent2": 1},
        wins={"agent1": 1, "agent2": 0},
        loses={"agent1": 0, "agent2": 1},
        ties={"agent1": 0, "agent2": 0},
        total_games=2,
        total_tournaments=1,
        std_dev={"agent1": 0, "agent2": 0},
    )


@pytest.fixture
def base_eval_config():
    return BaseEvaluatorConfig(force=True, verbose=True)


@pytest.fixture
def base_retrieval_eval_config(base_eval_config):
    base_config = base_eval_config.model_dump(exclude_unset=True)
    base_config["evaluator_name"] = RetrievalEvaluatorTypes.REASONER
    return BaseRetrievalEvaluatorConfig(
        **base_config,
    )


@pytest.fixture
def custom_answer_eval_config(base_eval_config, answer_eval_format):
    base_config = base_eval_config.model_dump(exclude_unset=True)
    base_config["evaluator_name"] = AnswerEvaluatorTypes.CUSTOM_PROMPT
    config = CustomPromptAnswerEvaluatorConfig(
        llm_response_schema=answer_eval_format,
        user_prompt=string_to_template(
            """ 
You are an useful assistant for evaluating the quality of the answers generated \
by RAG Agents. Given the following retrieved documents and a user query, evaluate \
the quality of the answers based on their quality, trustworthiness and originality. \
The last line of your answer must be a json object with the keys "quality", \
"trustworthiness" and originality, each of them with a single number between 0 and \
2, where 2 is the highest score on that aspect.
DOCUMENTS RETRIEVED:
{% for document in documents %}
{{document.text}}
{% endfor %}
User Query: {{ query.query }}

Agent answer: {{answer.text}}"""
        ),
        **base_config,
    )
    return config


@pytest.fixture
def expert_retrieval_eval_config(base_eval_config):
    base_eval_config = base_eval_config.model_dump(exclude_unset=True)
    base_eval_config["evaluator_name"] = AnswerEvaluatorTypes.DOMAIN_EXPERT
    return DomainExpertEvaluatorConfig(
        expert_in="Computer Science",
        domain_short="computer scientists",
        company="Zeta Alpha",
        extra_guidelines=["Super precise answers only!"],
        **base_eval_config,
    )


@pytest.fixture
def rdnam_config(base_eval_config):
    base_config = base_eval_config.model_dump(exclude_unset=True)
    base_config["query_file"] = "tests/data/rdnam_queries.csv"
    base_config["evaluator_name"] = RetrievalEvaluatorTypes.RDNAM
    return RDNAMEvaluatorConfig(
        annotator_role="You are a search quality rater evaluating the relevance of web pages. ",
        use_multiple_annotators=True,
        use_aspects=True,
        **base_config,
    )


@pytest.fixture
def custom_prompt_retrieval_eval_config(base_eval_config):
    base_eval_config = base_eval_config.model_dump(exclude_unset=True)
    base_eval_config["evaluator_name"] = RetrievalEvaluatorTypes.CUSTOM_PROMPT
    config = CustomPromptEvaluatorConfig(
        system_prompt="CUSTOM PROMPT SYSTEM PROMPT",
        user_prompt="query: {{ query.query }} doc: {{ document.text }}",
        **base_eval_config,
    )
    return config


@pytest.fixture
def llm_provider_mock(llm_provider_config):
    return MockLLMProvider(llm_provider_config)


@pytest.fixture
def llm_provider_mock_retrieval(llm_provider_config):
    mocked_answer = RetrievalAnswerEvaluatorFormat(
        reasoning="The document is very relevant",
        score=2,
    )
    LLM_response = LLMResponseType(
        raw_answer=mocked_answer.model_dump_json(),
        parsed_answer=mocked_answer,
    )
    provider = MockLLMProvider(llm_provider_config)
    provider.async_call_mocker = AsyncMock(side_effect=lambda *args, **kwargs: LLM_response)
    return provider


@pytest.fixture
def llm_provider_mock_rdnam(llm_provider_config):
    mocked_answer = RDNAMMultipleAnnotatorsAnswer(
        annotator_1=RDNAMAnswerEvaluatorFormat(intent_match=2, trustworthiness=1, overall=1),
        annotator_2=RDNAMAnswerEvaluatorFormat(intent_match=1, trustworthiness=1, overall=2),
        annotator_3=RDNAMAnswerEvaluatorFormat(intent_match=1, trustworthiness=1, overall=1),
        annotator_4=RDNAMAnswerEvaluatorFormat(intent_match=0, trustworthiness=0, overall=0),
        annotator_5=RDNAMAnswerEvaluatorFormat(intent_match=1, trustworthiness=1, overall=2),
    )
    LLM_response = LLMResponseType(raw_answer=mocked_answer.model_dump_json(), parsed_answer=mocked_answer)

    LLM_response = LLMResponseType(raw_answer=mocked_answer.model_dump_json(), parsed_answer=mocked_answer)
    provider = MockLLMProvider(llm_provider_config)

    def side_effect(*args, **kwargs):
        return LLM_response

    provider.async_call_mocker = AsyncMock(side_effect=side_effect)
    return provider


@pytest.fixture
def llm_provider_reasoner_mock(llm_provider_config):
    provider = MockLLMProvider(llm_provider_config)
    answer = LLMResponseType(
        raw_answer='{"reasoning": "The document is very relevant", "score": 2}',
        parsed_answer=RetrievalAnswerEvaluatorFormat(
            reasoning="The document is very relevant",
            score=2,
        ),
    )

    def side_effect(*args, **kwargs):
        return answer

    provider.async_call_mocker = AsyncMock(side_effect=side_effect)
    return provider


class AnswerFormat(BaseModel):
    keyA: str
    keyB: str


class AnswerEvaluatorFormat(BaseModel):
    quality: int
    trustworthiness: int
    originality: int


@pytest.fixture
def answer_format() -> Type[BaseModel]:
    return AnswerFormat


@pytest.fixture
def answer_eval_format() -> Type[BaseModel]:
    return AnswerEvaluatorFormat


@pytest.fixture
def base_answer_eval_config(base_eval_config, answer_eval_format):
    base_config = base_eval_config.model_dump(exclude_unset=True, exclude={"llm_response_schema"})
    base_config["system_prompt"] = "This is a system prompt"
    base_config["user_prompt"] = "Query: {{ query.query }}\nAnswer: {{ answer.text }}"
    base_config["llm_response_schema"] = answer_eval_format
    return BaseAnswerEvaluatorConfig(**base_config)


@pytest.fixture
def pairwise_answer_eval_config(base_answer_eval_config):
    base_config = base_answer_eval_config.model_dump(exclude_unset=True, exclude={"llm_response_schema"})
    base_config["pairwise"] = True
    base_config["evaluator_name"] = AnswerEvaluatorTypes.PAIRWISE
    base_config["user_prompt"] = """
        Query: {{ query.query }}
        [The Start of Assistant A's Answer]
        {{ game.agent_a_answer.text }}
        [The End of Assistant A's Answer]
        [The Start of Assistant B's Answer]
        {{ game.agent_b_answer.text }}
        [The End of Assistant B's Answer]
    """
    return PairwiseEvaluatorConfig(bidirectional=True, **base_config)


@pytest.fixture
def chat_pairwise_answer_eval_config(base_answer_eval_config):
    base_config = base_answer_eval_config.model_dump(
        exclude_unset=True, exclude={"user_prompt", "system_prompt", "llm_response_schema"}
    )
    base_config["pairwise"] = True
    base_config["evaluator_name"] = AnswerEvaluatorTypes.CHAT_PAIRWISE

    return PairwiseEvaluatorConfig(bidirectional=True, **base_config)


@pytest.fixture
def domain_expert_answer_eval_config(base_answer_eval_config):
    base_config = base_answer_eval_config.model_dump(
        exclude_unset=True, exclude={"user_prompt", "system_prompt", "llm_response_schema"}
    )
    base_config["pairwise"] = True
    base_config["expert_in"] = "Computer Science"
    base_config["include_annotations"] = True
    base_config["include_raw_documents"] = True
    base_config["evaluator_name"] = AnswerEvaluatorTypes.DOMAIN_EXPERT
    base_config["include_relevance_reasoning"] = False
    base_config["include_relevance_score"] = False
    return PairwiseDomainExpertEvaluatorConfig(**base_config)


@pytest.fixture
def few_shot_retrieval_eval_config(base_eval_config):
    base_eval_config = base_eval_config.model_dump(exclude_unset=True)
    base_eval_config["evaluator_name"] = RetrievalEvaluatorTypes.FEW_SHOT
    few_shot_samples = [
        FewShotExample(
            passage="Few shot example 1",
            query="Few shot query 1",
            relevance=2,
            reasoning="This is a good document",
        ),
        FewShotExample(
            passage="Few shot example 2",
            query="Few shot query 2",
            relevance=0,
            reasoning="This is a bad document",
        ),
    ]

    return FewShotEvaluatorConfig(
        system_prompt="System prompt",
        few_shots=few_shot_samples,
        **base_eval_config,
    )


@pytest.fixture
def llm_provider_answer_mock(llm_provider_config, answer_eval_format):
    provider = MockLLMProvider(llm_provider_config)
    mocked_answer = LLMResponseType(
        raw_answer='{"quality": 1, "trustworthiness": 0, "originality": 0}',
        parsed_answer=answer_eval_format(quality=1, trustworthiness=0, originality=0),
    )

    def side_effect(*args, **kwargs):
        return mocked_answer

    provider.async_call_mocker = AsyncMock(side_effect=side_effect)
    return provider



from ragelo.evaluators.retrieval_evaluators.base_retrieval_evaluator import (
    BaseRetrievalEvaluator,
    RetrievalEvaluatorFactory,
)
from ragelo.types.configurations import CustomPromptEvaluatorConfig
from ragelo.types.types import RetrievalEvaluatorTypes


@RetrievalEvaluatorFactory.register(RetrievalEvaluatorTypes.CUSTOM_PROMPT)
class CustomPromptEvaluator(BaseRetrievalEvaluator):
    config: CustomPromptEvaluatorConfig



from __future__ import annotations

import itertools
import random
from typing import Any, Callable, Set, Type, get_type_hints

from pydantic import BaseModel
from tenacity import RetryError

from ragelo.evaluators.base_evaluator import BaseEvaluator
from ragelo.llm_providers.base_llm_provider import BaseLLMProvider, get_llm_provider
from ragelo.logger import logger
from ragelo.types.configurations import BaseAnswerEvaluatorConfig, PairwiseEvaluatorConfig
from ragelo.types.evaluables import AgentAnswer, Document, Evaluable, PairwiseGame
from ragelo.types.experiment import Experiment
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.results import AnswerEvaluatorResult
from ragelo.types.types import AnswerEvaluatorTypes
from ragelo.utils import call_async_fn


class BaseAnswerEvaluator(BaseEvaluator):
    config: BaseAnswerEvaluatorConfig
    evaluable_name: str = "Agent Answer"
    _warned_queries: Set[str] = set()

    def evaluate(
        self,
        query: Query | str,
        answer: AgentAnswer | str | None = None,
        answer_a: AgentAnswer | str | None = None,
        answer_b: AgentAnswer | str | None = None,
        retrieved_documents: list[Document | str] | str | None = None,
        document_metadata: list[dict[str, Any]] | None = None,
        query_metadata: dict[str, Any] | None = None,
        answer_metadata: dict[str, Any] | None = None,
        answer_a_metadata: dict[str, Any] | None = None,
        answer_b_metadata: dict[str, Any] | None = None,
    ) -> AnswerEvaluatorResult:
        """
        Evaluates a single sample using an answer evaluator.
        Args:
            query(Query|str): The query to evaluate.
                If a string is provided, a Query object will be created with the provided query_metadata.
            answer(Optional[AgentAnswer|str]): The answer generated by an agent to evaluate.
                If a string is provided, the AgentAnswer object will be created with the provided answer_metadata
            answer_a(Optional[AgentAnswer|str]): The first answer generated by an agent to evaluate.
                Only used if the evaluator is pairwise.
            answer_b(Optional[AgentAnswer|str]): The second answer generated by an agent to evaluate.
                Only used if the evaluator is pairwise.
            retrieved_documents(Optional[list[str]|list[Document]]): The documents retrieved by the agent.
            document_metadata(Optional[list[dict[str, Any]]): The metadata of the documents retrieved by the agent.
            query_metadata(Optional[dict[str, Any]]): The metadata for the query.
            answer_metadata(Optional[dict[str, Any]]): The metadata for the answer.
            answer_a_metadata(Optional[dict[str, Any]]): The metadata for the first answer.
                Only used if the evaluator is pairwise.
            answer_b_metadata(Optional[dict[str, Any]]): The metadata for the second answer.
                Only used if the evaluator is pairwise.
        Returns:
            AnswerEvaluatorResult: The result of the evaluation.
        """
        query = Query.assemble_query(query, query_metadata)

        if isinstance(retrieved_documents, str):
            retrieved_documents = [retrieved_documents]

        if retrieved_documents:
            retrieved_and_assembled_docs = Document.assemble_documents(
                retrieved_documents, query.qid, document_metadata
            )
            query.retrieved_docs = retrieved_and_assembled_docs

        agent: str | tuple[str, str]
        evaluable: AgentAnswer | PairwiseGame

        if self.config.pairwise:
            if not answer_a or not answer_b:
                raise ValueError("Pairwise evaluations require two answers")
            answer_a = AgentAnswer.assemble_answer(answer_a, query.qid, metadata=answer_a_metadata)
            answer_b = AgentAnswer.assemble_answer(answer_b, query.qid, metadata=answer_b_metadata)
            agent = (answer_a.agent, answer_b.agent)
            evaluable = PairwiseGame(
                qid=query.qid,
                agent_a_answer=answer_a,
                agent_b_answer=answer_b,
            )
        else:
            if not answer:
                raise ValueError("Pointwise evaluations require an answer")
            evaluable = AgentAnswer.assemble_answer(answer, query.qid, metadata=answer_metadata)
            agent = evaluable.agent
        result = call_async_fn(self.evaluate_async, (query, evaluable))

        if result.exception or result.raw_answer is None or result.answer is None:
            raise ValueError(
                f"Failed to evaluate qid: {query.qid} agent(s): {agent}",
                f"Exception: {result.exception}",
            )
        return result

    def evaluate_experiment(self, experiment: Experiment, n_threads: int | None = None):
        """
        Trigger the evaluation of all the queries in the experiment.
        The evaluation is done in asynchronously with the number of threads defined in config.n_processes parameter.
        This can be overwritten by the n_threads parameter.

        Args:
            experiment(Experiment): The experiment to evaluate.
            n_threads(int): The number of threads to use for the evaluation.
                If None, the number of threads defined in the config will be used.
        """
        if self.config.pairwise:
            self.__add_pairwise_games(experiment)
        super().evaluate_experiment(experiment, n_threads)

    async def evaluate_async(self, eval_sample: tuple[Query, Evaluable]) -> AnswerEvaluatorResult:
        """
        Evaluates a single sample (either an answer or a pairwise game) asynchronously.
        Args:
            eval_sample (tuple[Query, Evaluable]): The query and evaluable to evaluate.
        """
        agent: str | tuple[str, str]
        query, evaluable = eval_sample

        if not isinstance(evaluable, AgentAnswer) and not isinstance(evaluable, PairwiseGame):
            type_name = type(evaluable).__name__
            raise ValueError(f"can't evaluate a {type_name} in an Answer Evaluator")

        if evaluable.evaluation is not None and not self.config.force:
            if isinstance(evaluable, AgentAnswer):
                return AnswerEvaluatorResult(
                    qid=query.qid,
                    agent=evaluable.agent,
                    raw_answer=evaluable.evaluation.raw_answer,
                    answer=evaluable.evaluation.answer,
                    pairwise=False,
                )
            return AnswerEvaluatorResult(
                qid=query.qid,
                agent_a=evaluable.agent_a_answer.agent,
                agent_b=evaluable.agent_b_answer.agent,
                raw_answer=evaluable.evaluation.raw_answer,
                answer=evaluable.evaluation.answer,
                pairwise=True,
            )

        if isinstance(evaluable, AgentAnswer):
            agent = evaluable.agent
            prompt = self._build_message(query, evaluable)
        elif isinstance(evaluable, PairwiseGame):
            agent = (evaluable.agent_a_answer.agent, evaluable.agent_b_answer.agent)
            prompt = self._build_message_pairwise(query, evaluable)
        else:
            raise ValueError(f"Unknown evaluable type {type(evaluable)}")

        exc = None
        try:
            llm_response = await self.llm_provider.call_async(
                input=prompt,
                response_schema=self.config.llm_response_schema,
            )
            llm_response = self._process_answer(llm_response)
        except ValueError as e:
            logger.warning(
                f"Failed to PARSE answer for qid: {query.qid} agent(s): {agent}\nRaw answer: {llm_response.raw_answer}"
            )
            exc = str(e)
        except Exception as e:
            logger.warning(f"Failed to FETCH answers for qid: {query.qid}")
            logger.warning(f"agent(s): {agent}")
            if isinstance(e, RetryError):
                exc = str(e.last_attempt.exception())
            else:
                exc = str(e)
        if isinstance(evaluable, AgentAnswer):
            return AnswerEvaluatorResult(
                qid=query.qid,
                agent=evaluable.agent,
                raw_answer=llm_response.raw_answer,
                answer=llm_response.parsed_answer,
                pairwise=False,
                exception=exc,
            )
        assert isinstance(evaluable, PairwiseGame)
        return AnswerEvaluatorResult(
            qid=query.qid,
            agent_a=evaluable.agent_a_answer.agent,
            agent_b=evaluable.agent_b_answer.agent,
            raw_answer=llm_response.raw_answer,
            answer=llm_response.parsed_answer,
            pairwise=True,
            exception=exc,
        )

    def _get_tuples_to_evaluate(self, experiment: Experiment) -> list[tuple[Query, Evaluable]]:
        """
        Creates the list of pairs (query, evaluable) to evaluate
        """
        tuples_to_eval: list[tuple[Query, Evaluable]] = []
        all_tuples = 0
        missing_evaluations = 0
        for q in experiment:
            if self.config.pairwise:
                for g in q.pairwise_games:
                    all_tuples += 1
                    tuples_to_eval.append((q, g))
                    if g.evaluation is None:
                        missing_evaluations += 1

            else:
                for a in q.answers.values():
                    all_tuples += 1
                    tuples_to_eval.append((q, a))
                    if a.evaluation is None:
                        missing_evaluations += 1

        if missing_evaluations == 0 and not self.config.force:
            logger.info(
                f"All {all_tuples} answers are already evaluated.\n"
                "If you want to re-evaluate them, use the --force flag"
            )

        return tuples_to_eval

    def _build_message(self, query: Query, answer: AgentAnswer) -> LLMInputPrompt:
        """Builds the message to send to the LLM evaluator"""
        documents = self._filter_documents(query)
        context = {"query": query, "answer": answer, "documents": documents}
        user_message = self.user_prompt.render(**context)
        system_prompt = self.system_prompt.render(**context) if self.system_prompt else None
        return LLMInputPrompt(
            system_prompt=system_prompt,
            user_message=user_message,
        )

    def _build_message_pairwise(self, query: Query, game: PairwiseGame) -> LLMInputPrompt:
        """Builds the message to send to the LLM evaluator based on a pairwise game"""
        documents = self._filter_documents(query)
        context = {"query": query, "game": game, "documents": documents}
        user_message = self.user_prompt.render(**context)
        system_prompt = self.system_prompt.render(**context) if self.system_prompt else None
        return LLMInputPrompt(
            system_prompt=system_prompt,
            user_message=user_message,
        )

    @classmethod
    def from_config(cls, config: BaseAnswerEvaluatorConfig, llm_provider: BaseLLMProvider):
        return cls(config, llm_provider)

    @classmethod
    def get_config_class(cls) -> Type[BaseAnswerEvaluatorConfig]:
        return get_type_hints(cls)["config"]

    def __add_pairwise_games(self, experiment: Experiment):
        if not self.config.pairwise:
            return
        if not isinstance(self.config, PairwiseEvaluatorConfig):
            raise ValueError("Trying to add pairwise games to a non-pairwise evaluator")
        for query in experiment:
            query_agents = list(query.answers.keys())
            pairs = list(itertools.combinations(query_agents, 2))
            if self.config.bidirectional:
                pairs += [(b, a) for a, b in pairs]
            random.shuffle(pairs)

            # Filter out games that already exist
            existing_games = {(a.agent_a_answer.agent, a.agent_b_answer.agent) for a in query.pairwise_games}
            games = [g for g in pairs if g not in existing_games]

            games_to_add = min(self.config.n_games_per_query - len(existing_games), len(games))
            games = games[:games_to_add]
            for agent_a, agent_b in games:
                query.pairwise_games.append(
                    PairwiseGame(
                        qid=query.qid,
                        agent_a_answer=query.answers[agent_a],
                        agent_b_answer=query.answers[agent_b],
                    )
                )

    def _filter_documents(self, query: Query) -> list[Document]:
        documents = []
        for did, d in query.retrieved_docs.items():
            if self.config.document_relevance_threshold is not None:
                # Skip documents with relevance below the threshold
                if d.evaluation is None:
                    continue
                # check if evaluation.answer is an integer or a string that can be converted to an integer
                score = d.evaluation.answer
                if isinstance(d.evaluation.answer, dict):
                    score = d.evaluation.answer["score"]
                elif isinstance(d.evaluation.answer, BaseModel):
                    score = d.evaluation.answer.score  # type: ignore
                else:
                    score = d.evaluation.answer

                if isinstance(score, str) and score.isdigit():
                    score = int(score)
                if not isinstance(score, (int, float)):
                    continue
                if score < self.config.document_relevance_threshold:
                    continue
            if self.config.document_filter is not None and not self.config.document_filter(d):
                continue
            documents.append(d)
        if len(documents) == 0:
            if query.qid not in self._warned_queries:
                logger.warning(
                    f"No relevant documents were retrieved for the query {query.qid}. "
                    "No documents will be provided to the Answer Evaluator."
                )
                self._warned_queries.add(query.qid)
        return documents


class AnswerEvaluatorFactory:
    registry: dict[AnswerEvaluatorTypes, Type[BaseAnswerEvaluator]] = {}

    @classmethod
    def register(cls, name: AnswerEvaluatorTypes) -> Callable:
        def inner_wrapper(wrapped_class: Type[BaseAnswerEvaluator]):
            if name in cls.registry:
                logger.warning(f"Overwriting {name} in registry")
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(
        cls,
        evaluator_name: AnswerEvaluatorTypes,
        llm_provider: BaseLLMProvider | str,
        config: BaseAnswerEvaluatorConfig | None = None,
        **kwargs,
    ) -> BaseAnswerEvaluator:
        if evaluator_name not in cls.registry:
            raise ValueError(
                f"Unknown answer evaluator {evaluator_name}\nValid options are {list(cls.registry.keys())}"
            )
        if isinstance(llm_provider, str):
            llm_provider_instance = get_llm_provider(llm_provider, **kwargs)
        else:
            llm_provider_instance = llm_provider
        if config is None:
            class_ = cls.registry[evaluator_name]
            type_config = class_.get_config_class()
            valid_keys = [field for field in type_config.model_fields]
            valid_args = {k: v for k, v in kwargs.items() if k in valid_keys}
            required_fields = [arg for arg, info in type_config.model_fields.items() if info.is_required()]
            for field in required_fields:
                if field not in valid_args:
                    raise ValueError(f"Required argument {field} for evaluator {evaluator_name} not provided")
            config = type_config(**valid_args)
        return cls.registry[evaluator_name].from_config(config, llm_provider_instance)


def get_answer_evaluator(
    evaluator_name: AnswerEvaluatorTypes | str | None = None,
    llm_provider: BaseLLMProvider | str = "openai",
    config: BaseAnswerEvaluatorConfig | None = None,
    **kwargs,
) -> BaseAnswerEvaluator:
    """Gets an answer evaluator
    Args:
        evaluator_name (AnswerEvaluatorTypes | str | None): The name of the answer evaluator to get.
        llm_provider (BaseLLMProvider | str): The LLM provider to use for the evaluation.
        config (BaseAnswerEvaluatorConfig | None): The configuration for the answer evaluator.
        **kwargs: Additional arguments to pass to the answer evaluator.
    Returns:
        BaseAnswerEvaluator: The answer evaluator.
    """
    if evaluator_name is None:
        # get the name from the config
        if config is None:
            raise ValueError("Either the evaluator_name or a config object must be provided")
        evaluator_name = config.evaluator_name
    if isinstance(evaluator_name, str):
        try:
            evaluator_name = AnswerEvaluatorTypes(evaluator_name)
        except ValueError:
            raise ValueError(f"Unknown answer evaluator {evaluator_name}")
    if evaluator_name is None:
        raise ValueError("The evaluator_name must be provided")
    return AnswerEvaluatorFactory.create(
        evaluator_name,
        llm_provider=llm_provider,
        config=config,
        **kwargs,
    )



"""Answer Evaluator with a domain expert persona"""

from ragelo.evaluators.answer_evaluators.base_answer_evaluator import AnswerEvaluatorFactory
from ragelo.evaluators.answer_evaluators.pairwise_evaluator import PairwiseAnswerEvaluator
from ragelo.types.configurations import PairwiseDomainExpertEvaluatorConfig
from ragelo.types.evaluables import PairwiseGame
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.types import AnswerEvaluatorTypes
from ragelo.utils import string_to_template


@AnswerEvaluatorFactory.register(AnswerEvaluatorTypes.DOMAIN_EXPERT)
class PairwiseDomainExpertEvaluator(PairwiseAnswerEvaluator):
    config: PairwiseDomainExpertEvaluatorConfig
    system_prompt = string_to_template("""
        You are a domain expert in {{ expert_in }}.{% if company %} You work for {{ company }}.{% endif %} You are tasked with evaluating the quality of the responses provided by two AI assistants that were tasked with answering a user's question based on a set of documents retrieved by a search engine.
        When available, answers will cite specific documents by placing their IDs into square brackets.
        {%- if doc and (annotation or reasoning) %}
        You will be provided with the text of each reference document and its relevance evaluation.
        {%- elif doc %}
        You will be provided with the text of each reference document.
        {%- elif annotation or reasoning %}
        For each cited document, you will be provided with its relevance evaluation.
        {%- endif %}
        {%- if not reasoning %}
        Each document is scored in a scale of 0 to 2, where:
            - 0: the document is not relevant to the query
            - 1: the document is somewhat relevant to the query
            - 2: the document is highly relevant to the query
        {%- endif %}

        You should choose the assistant that best answers the user's question.

        ## Evaluation Guidelines
        Your evaluation should consider factors such as {{ factors}}. 
        You should follow the following guidelines:
        - Details are only useful if they answer the user question.
        - If an answer contains non-relevant details, it should not be preferred over one that only use relevant information.
        - Avoid any position biases and ensure that the order in which the responses were presented does not influence your decision.
        - Do not allow the length of the responses to influence your evaluation.
        - Be as objective as possible.
        - Remember that you are in expert in {{ expert_in }}. Make your judgement accordingly.

        ## Workflow
        First, you should analyze each of the two answers, explaining whether or not each of them correctly answers the user's question, based on the relevant documents retrieved and your expertise.
        Then, you should compare the two responses and provide a short explanation on their differences, explaining in which aspects each answer is better or worst than the other. 
        After providing your explanation, output your final verdict by strictly following his format: "A" if assistant A is better, "B" if assistant B is better, or "C" for a tie.""")

    def _build_message_pairwise(self, query: Query, game: PairwiseGame) -> LLMInputPrompt:
        documents = self._filter_documents(query)
        context = {
            "factors": self.config.factors,
            "query": query,
            "documents": documents,
            "game": game,
            "doc": self.config.include_raw_documents,
            "annotation": self.config.include_relevance_score,
            "expert_in": self.config.expert_in,
            "company": self.config.company,
            "reasoning": self.config.include_relevance_reasoning,
        }

        return LLMInputPrompt(
            system_prompt=self.system_prompt.render(**context),
            user_message=self.user_prompt.render(**context),
        )



from __future__ import annotations

import random

import numpy as np

from ragelo.agent_rankers.base_agent_ranker import AgentRanker, AgentRankerFactory
from ragelo.logger import logger
from ragelo.types.configurations import EloAgentRankerConfig
from ragelo.types.experiment import Experiment
from ragelo.types.results import EloTournamentResult
from ragelo.types.types import AgentRankerTypes


@AgentRankerFactory.register(AgentRankerTypes.ELO)
class EloRanker(AgentRanker):
    name: str = "Elo Agent Ranker"
    config: EloAgentRankerConfig

    def __init__(
        self,
        config: EloAgentRankerConfig,
    ):
        super().__init__(config)
        self.score_map = config.score_mapping

        self.agents_scores: dict[str, float] = {}
        self.wins: dict[str, int] = {}
        self.losses: dict[str, int] = {}
        self.ties: dict[str, int] = {}
        self.total_games: int = 0
        self.games_played: dict[str, int] = {}
        self.computed: bool = False
        self.initial_score: int = self.config.initial_score
        self.k: int = self.config.elo_k
        self.std_dev: dict[str, float] = {}

    def run(self, experiment: Experiment) -> EloTournamentResult:
        """Compute score for each agent"""
        self.evaluations = self._flatten_evaluations(experiment)
        agent_scores: dict[str, list[int]] = {}
        for _ in range(self.config.tournaments):
            results = self.run_tournament()
            for agent, score in results.items():
                agent_scores[agent] = agent_scores.get(agent, []) + [score]
        for a in agent_scores:
            self.std_dev[a] = float(np.std(agent_scores[a]))
            self.agents_scores[a] = float(np.mean(agent_scores[a]))

        result = EloTournamentResult(
            agents=list(self.agents_scores.keys()),
            scores=self.agents_scores,
            games_played=self.games_played,
            wins=self.wins,
            loses=self.losses,
            ties=self.ties,
            std_dev=self.std_dev,
            total_games=self.total_games,
            total_tournaments=self.config.tournaments,
        )
        experiment.add_evaluation(result, should_print=True)
        return result

    def get_agents_ratings(self):
        return self.agents_scores

    def get_ranked_agents(self) -> list[tuple[str, float]]:
        ranking = sorted(self.get_agents_ratings().items(), key=lambda x: x[1], reverse=True)
        return [(agent, rating) for agent, rating in ranking]

    def run_tournament(self) -> dict[str, int]:
        agents_scores: dict[str, int] = {}
        games: list[tuple[str, str, float]] = []
        for agent_a, agent_b, score in self.evaluations:
            score_val = self.score_map[score]
            games.append((agent_a, agent_b, score_val))
        random.shuffle(games)
        for agent_a, agent_b, score_val in games:
            if self.config.verbose:
                logger.info(f"Game: {agent_a} vs {agent_b} -> {score_val}")
            if score_val == 1:
                self.wins[agent_a] = self.wins.get(agent_a, 0) + 1
                self.losses[agent_b] = self.losses.get(agent_b, 0) + 1
            elif score_val == 0:
                self.wins[agent_b] = self.wins.get(agent_b, 0) + 1
                self.losses[agent_a] = self.losses.get(agent_a, 0) + 1
            else:
                self.ties[agent_a] = self.ties.get(agent_a, 0) + 1
                self.ties[agent_b] = self.ties.get(agent_b, 0) + 1
            agent_a_rating = agents_scores.get(agent_a, self.initial_score)
            agent_b_rating = agents_scores.get(agent_b, self.initial_score)

            expected_score = 1 / (1 + 10 ** ((agent_a_rating - agent_b_rating) / 400))
            agents_scores[agent_a] = int(agent_a_rating + self.k * (score_val - expected_score))
            agents_scores[agent_b] = int(agent_b_rating + self.k * ((1 - score_val) - (1 - expected_score)))
            self.total_games += 1
            self.games_played[agent_a] = self.games_played.get(agent_a, 0) + 1
            self.games_played[agent_b] = self.games_played.get(agent_b, 0) + 1

        self.computed = True
        return agents_scores



from enum import Enum


class StrEnum(str, Enum):
    # Implementing __str__ to allow for easy conversion to string
    # e.g. str(RetrievalEvaluatorTypes.CUSTOM_PROMPT) -> "custom_prompt"
    # otherwise, it's "RetrievalEvaluatorTypes.CUSTOM_PROMPT"
    def __str__(self):
        return self.value


class RetrievalEvaluatorTypes(StrEnum):
    """Enum that contains the names of the available retrieval evaluators"""

    CUSTOM_PROMPT = "custom_prompt"
    DOMAIN_EXPERT = "domain_expert"
    FEW_SHOT = "few_shot"
    RDNAM = "RDNAM"
    REASONER = "reasoner"


class LLMProviderTypes(StrEnum):
    """Enum that contains the names of the available LLM providers"""

    OPENAI = "openai"
    OLLAMA = "ollama"


class AnswerEvaluatorTypes(StrEnum):
    """Enum that contains the names of the available answer evaluators"""

    PAIRWISE = "pairwise"
    CUSTOM_PAIRWISE = "custom_pairwise"
    CUSTOM_PROMPT = "custom_prompt"
    DOMAIN_EXPERT = "domain_expert"
    CHAT_PAIRWISE = "chat_pairwise"


class AgentRankerTypes(StrEnum):
    ELO = "elo"



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import Any, Type, get_type_hints

from pydantic import BaseModel

from ragelo.types.configurations import LLMProviderConfig
from ragelo.types.formats import LLMInputPrompt, LLMResponseType
from ragelo.types.types import LLMProviderTypes
from ragelo.utils import call_async_fn


class BaseLLMProvider(ABC):
    config: LLMProviderConfig
    api_key_env_var: str = "OPENAI_API_KEY"

    def __init__(self, config: LLMProviderConfig):
        self.config = config

    def __call__(
        self,
        input: LLMInputPrompt,
        response_schema: Type[BaseModel] | dict[str, Any] | None = None,
    ) -> LLMResponseType:
        """Submits a single query-document pair to the LLM and returns the answer."""
        return call_async_fn(self.call_async, input, response_schema)

    @abstractmethod
    async def call_async(
        self,
        input: LLMInputPrompt,
        response_schema: Type[BaseModel] | dict[str, Any] | None = None,
    ) -> LLMResponseType:
        """Submits a single query-document pair to the LLM and returns the answer."""
        raise NotImplementedError

    @classmethod
    def from_config(
        cls,
        config: LLMProviderConfig,
    ) -> "BaseLLMProvider":
        """Inits the LLM provider from a credentials file."""
        return cls(config)

    @classmethod
    def get_config_class(cls) -> Type[LLMProviderConfig]:
        return get_type_hints(cls)["config"]


class LLMProviderFactory:
    registry: dict[LLMProviderTypes, Type[BaseLLMProvider]] = {}

    @classmethod
    def register(cls, name: LLMProviderTypes):
        """Registers a new LLM provider"""

        def inner_wrapper(
            wrapped_class: Type[BaseLLMProvider],
        ) -> Type[BaseLLMProvider]:
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(
        cls,
        name: LLMProviderTypes,
        config: LLMProviderConfig | None = None,
        **kwargs,
    ) -> BaseLLMProvider:
        """Creates a new LLM provider"""
        if name not in cls.registry:
            raise ValueError(f"LLM provider {name} not found")
        if config is None:
            class_ = cls.registry[name]
            type_config = class_.get_config_class()
            valid_keys = [field for field in type_config.model_fields]
            if "api_key" not in kwargs:
                api_key = os.environ.get(class_.api_key_env_var)
                if not api_key:
                    # Check if the key is actually required
                    api_key_field = type_config.model_fields["api_key"]
                    is_required = api_key_field.is_required()
                    if is_required:
                        raise ValueError(f"API key not found in environment variable {class_.api_key_env_var}")
                    else:
                        api_key = api_key_field.default
                kwargs["api_key"] = api_key
            valid_args = {k: v for k, v in kwargs.items() if k in valid_keys}
            config = type_config(**valid_args)
        return cls.registry[name].from_config(config)


def get_llm_provider(
    name: LLMProviderTypes | str,
    config: LLMProviderConfig | None = None,
    **kwargs,
) -> BaseLLMProvider:
    """Creates a new LLM provider"""
    if isinstance(name, str):
        name = LLMProviderTypes(name)
    return LLMProviderFactory.create(name, config, **kwargs)



from ragelo.types.configurations.agent_ranker_configs import AgentRankerConfig, EloAgentRankerConfig
from ragelo.types.configurations.answer_evaluator_configs import (
    BaseAnswerEvaluatorConfig,
    CustomPairwiseEvaluatorConfig,
    CustomPromptAnswerEvaluatorConfig,
    PairwiseDomainExpertEvaluatorConfig,
    PairwiseEvaluatorConfig,
)
from ragelo.types.configurations.base_configs import BaseConfig, BaseEvaluatorConfig
from ragelo.types.configurations.cli_configs import CLIConfig
from ragelo.types.configurations.llm_provider_configs import (
    LLMProviderConfig,
    OllamaConfiguration,
    OpenAIConfiguration,
)
from ragelo.types.configurations.retrieval_evaluator_configs import (
    BaseRetrievalEvaluatorConfig,
    CustomPromptEvaluatorConfig,
    DomainExpertEvaluatorConfig,
    FewShotEvaluatorConfig,
    RDNAMEvaluatorConfig,
    ReasonerEvaluatorConfig,
)

__all__ = [
    "CLIConfig",
    "BaseConfig",
    "BaseEvaluatorConfig",
    "LLMProviderConfig",
    "OpenAIConfiguration",
    "OllamaConfiguration",
    "AgentRankerConfig",
    "EloAgentRankerConfig",
    "BaseAnswerEvaluatorConfig",
    "CustomPairwiseEvaluatorConfig",
    "CustomPromptAnswerEvaluatorConfig",
    "PairwiseEvaluatorConfig",
    "PairwiseDomainExpertEvaluatorConfig",
    "BaseRetrievalEvaluatorConfig",
    "FewShotEvaluatorConfig",
    "RDNAMEvaluatorConfig",
    "ReasonerEvaluatorConfig",
    "CustomPromptEvaluatorConfig",
    "DomainExpertEvaluatorConfig",
]



from ragelo.evaluators.answer_evaluators.base_answer_evaluator import AnswerEvaluatorFactory, BaseAnswerEvaluator
from ragelo.types.configurations import PairwiseEvaluatorConfig
from ragelo.types.evaluables import PairwiseGame
from ragelo.types.formats import LLMInputPrompt
from ragelo.types.query import Query
from ragelo.types.types import AnswerEvaluatorTypes
from ragelo.utils import string_to_template


@AnswerEvaluatorFactory.register(AnswerEvaluatorTypes.PAIRWISE)
class PairwiseAnswerEvaluator(BaseAnswerEvaluator):
    """An evaluator that evaluates RAG-based answers pairwise, with document reasoning and citations."""

    config: PairwiseEvaluatorConfig
    user_prompt_document = "[{did}] {doc}"
    user_prompt_annotation = "[{did}] {annotation}"
    user_prompt_document_and_annotation = "[{{ d.did }}] Content: {{ d.text }}\n Evaluation: {{ d.evaluation.answer }}"

    system_prompt = string_to_template("""
        Please act as an impartial judge and evaluate the quality of the responses provided by two AI assistants tasked to answer the question of a user, based on a set of documents retrieved by a search engine that may or may not be relevant to the question. 
        When available, answers will cite specific documents by placing their IDs into square brackets.
        {%- if doc and (annotation or reasoning) %}
        You will be provided with the text of each reference document and its relevance evaluation.
        {%- elif doc %}
        You will be provided with the text of each reference document.
        {%- elif annotation or reasoning %}
        For each cited document, you will be provided with its relevance evaluation.
        {%- endif %}
        {%- if not reasoning %}
        Each document is scored in a scale of 0 to 2, where:
            - 0: the document is not relevant to the query
            - 1: the document is somewhat relevant to the query
            - 2: the document is highly relevant to the query
        {%- endif %}

        You should choose the assistant that best answers the user's question.

        ## Evaluation Guidelines
        - Your evaluation should consider factors such as {{ factors }}.
        - Details are only useful if they answer the user question.
        - If an answer contains non-relevant details, it should not be preferred over one that only use relevant information.
        - Avoid any position biases and ensure that the order in which the responses were presented does not influence your decision.
        - Do not allow the length of the responses to influence your evaluation.
        - Be as objective as possible.

        ## Workflow
        First, you should analyze each of the two answers, explaining whether or not each of them correctly answers the user's question, based on the relevant documents retrieved. 
        Then, you should compare the two responses and provide a short explanation on their differences, explaining in which aspects each answer is better or worst than the other. 
        After providing your explanation, output your final verdict by strictly following his format: "A" if assistant A is better, "B" if assistant B is better, or "C" for a tie.""")

    user_prompt = string_to_template("""
        [User Question]
        {{ query.query }}
        {%- if documents %}
        [Reference Documents]
        {%- for d in documents %}
        {%- if doc and (annotation or reasoning) %}
            Document ID: [{{ d.did }}]
            Content: {{ d.text }}
            Relevance: {% if reasoning %} {{ d.evaluation.answer.reasoning }} {% else %} {{ d.evaluation.answer.score }} {% endif %}
        ------------------
        {%- elif doc %}
            [{{ d.did }}]: {{ d.text }}
        {%- elif annotation or reasoning %}
            [{{d.did }}] {% if reasoning %} {{ d.evaluation.answer.reasoning }} {% else %} {{ d.evaluation.answer.score }} {% endif %}"
        {% endif -%}
        {% endfor %}
        {% endif -%}

        [The Start of Assistant A's Answer]
            {{ game.agent_a_answer.text }}
        [The End of Assistant A's Answer]

        [The Start of Assistant B's Answer]
            {{ game.agent_b_answer.text }}
        [The End of Assistant B's Answer]""")

    def _build_message_pairwise(self, query: Query, game: PairwiseGame) -> LLMInputPrompt:
        documents = self._filter_documents(query)
        context = {
            "factors": self.config.factors,
            "query": query,
            "documents": documents,
            "game": game,
            "doc": self.config.include_raw_documents,
            "annotation": self.config.include_relevance_score,
            "reasoning": self.config.include_relevance_reasoning,
        }
        return LLMInputPrompt(
            system_message=self.system_prompt.render(**context),
            user_message=self.user_prompt.render(**context),
        )



from ragelo.llm_providers.base_llm_provider import BaseLLMProvider, LLMProviderFactory, get_llm_provider
from ragelo.llm_providers.ollama_client import OllamaProvider
from ragelo.llm_providers.openai_client import OpenAIProvider

__all__ = [
    "BaseLLMProvider",
    "LLMProviderFactory",
    "OllamaProvider",
    "OpenAIProvider",
    "get_llm_provider",
]



from __future__ import annotations

from pydantic import Field

from ragelo.types.configurations.base_configs import BaseConfig


class AgentRankerConfig(BaseConfig):
    ranker_name: str = Field(default="elo", description="The name of the agent ranker")
    verbose: bool = Field(default=False, description="Whether to print each game played")


class EloAgentRankerConfig(AgentRankerConfig):
    elo_k: int = Field(default=32, description="The K factor for the Elo ranking algorithm")
    initial_score: int = Field(default=1000, description="The initial score for each agent")
    tournaments: int = Field(default=10, description="The number of Elo tournaments to play")
    score_mapping: dict[str, float] = Field(
        default={"A": 1, "B": 0, "C": 0.5},
        description="The mapping of the evaluation answers to scores",
    )



import logging

import typer

from ragelo import Experiment, get_llm_provider, get_retrieval_evaluator
from ragelo.cli.args import get_params_from_function
from ragelo.cli.utils import get_path
from ragelo.types.configurations.cli_configs import (
    CLIDomainExpertEvaluatorConfig,
    CLIRDNAMEvaluatorConfig,
    CLIReasonerEvaluatorConfig,
)
from ragelo.types.types import RetrievalEvaluatorTypes

typer.main.get_params_from_function = get_params_from_function  # type: ignore


app = typer.Typer()


@app.command()
def domain_expert(config: CLIDomainExpertEvaluatorConfig = CLIDomainExpertEvaluatorConfig(), **kwargs):
    """Evaluator with a domain expert persona.

    This Retrieval Evaluator evaluates the relevance of documents submitted by
    a user that is an expert in a specific domain. For instance,
    to evaluate the documents retrieved to queries submitted by a Chemical
    Engineer that works at ChemCorp Inc:

    ragelo retrieval_evaluator domain_expert queries.csv documents.csv "Chemical Engineering" --company "ChemCorp Inc."

    """
    logging.getLogger("ragelo").setLevel(logging.INFO)
    kwargs.pop("llm_response_schema", None)

    config = CLIDomainExpertEvaluatorConfig(**kwargs)
    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    kwargs = config.model_dump()

    evaluator = get_retrieval_evaluator(
        RetrievalEvaluatorTypes.DOMAIN_EXPERT, config=config, llm_provider=llm_provider
    )
    evaluator.evaluate_experiment(experiment)
    experiment.save(output_file)


@app.command()
def reasoner(
    config: CLIReasonerEvaluatorConfig = CLIReasonerEvaluatorConfig(),
    **kwargs,
):
    """
    A document Evaluator that only outputs the reasoning for why a document is relevant.
    """
    logging.getLogger("ragelo").setLevel(logging.INFO)

    kwargs.pop("llm_response_schema", None)

    config = CLIReasonerEvaluatorConfig(**kwargs)

    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    kwargs = config.model_dump()

    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    evaluator = get_retrieval_evaluator(RetrievalEvaluatorTypes.REASONER, config=config, llm_provider=llm_provider)
    evaluator.evaluate_experiment(experiment)
    experiment.save(output_file)


@app.command()
def rdnam(config: CLIRDNAMEvaluatorConfig = CLIRDNAMEvaluatorConfig(), **kwargs):
    """
    Evaluator based on the paper by Thomas, Spielman, Craswell and Mitra:
    Large language models can accurately predict searcher preferences.
    """
    logging.getLogger("ragelo").setLevel(logging.INFO)

    kwargs.pop("llm_response_schema", None)

    config = CLIRDNAMEvaluatorConfig(**kwargs)
    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    kwargs = config.model_dump()
    kwargs.pop("llm_response_schema")

    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    evaluator = get_retrieval_evaluator(RetrievalEvaluatorTypes.RDNAM, config=config, llm_provider=llm_provider)
    evaluator.evaluate_experiment(experiment)
    experiment.save(output_file)


if __name__ == "__main__":
    app()



import os

import pytest
from typer.testing import CliRunner

from ragelo.cli.cli import app

runner = CliRunner()


@pytest.mark.requires_openai
def test_run_all_cli():
    if os.path.exists("tests/data/output.json"):
        os.remove("tests/data/output.json")
    if os.path.exists("tests/data/output_results.jsonl"):
        os.remove("tests/data/output_results.jsonl")
    result = runner.invoke(
        app,
        [
            "run-all",
            "queries.csv",
            "documents.csv",
            "answers.csv",
            "--verbose",
            "--data-dir",
            "tests/data/",
            "--experiment-name",
            "test",
            "--output-file",
            "output.json",
            "--force",
        ],
    )
    assert result.exit_code == 0
    assert "Agents Elo Ratings" in result.stdout
    assert result.stdout.startswith("Creating a cache file for the experiment's evaluations")
    assert "🔎 Query ID: 0\n📜 Document ID: 0" in result.stdout
    assert "🔎 Query ID: 0\n agent1              🆚   agent2\nParsed Answer:" in result.stdout
    assert "Total evaluations: 4" in result.stdout
    assert "Total evaluations: 2" in result.stdout
    assert "Evaluating Retrieved documents" in result.stdout
    assert "Evaluating Agent Answers" in result.stdout
    assert len(result.stdout.split("✅ Done!")) == 3
    assert os.path.exists("tests/data/output.json")
    assert os.path.exists("tests/data/output_results.jsonl")
    assert not os.path.exists("ragelo_cache/test.json")
    os.remove("tests/data/output.json")
    os.remove("tests/data/output_results.jsonl")


@pytest.mark.requires_openai
def test_run_reasoner_cli():
    if os.path.exists("tests/data/test-output.json"):
        os.remove("tests/data/test-output.json")
    result = runner.invoke(
        app,
        [
            "retrieval-evaluator",
            "reasoner",
            "queries.csv",
            "documents.csv",
            "--verbose",
            "--data-dir",
            "tests/data/",
            "--experiment-name",
            "test",
            "--output-file",
            "test-output.json",
            "--no-save-results",
            "--force",
        ],
    )
    assert result.exit_code == 0
    assert "✅ Done!" in result.stdout
    assert result.stdout.startswith("Loaded 2 queries from")
    assert "🔎 Query ID: 0\n📜 Document ID: 0" in result.stdout
    assert "Total evaluations: 4" in result.stdout
    assert os.path.exists("tests/data/test-output.json")
    os.remove("tests/data/test-output.json")


@pytest.mark.requires_openai
def test_run_answer_cli():
    result = runner.invoke(
        app,
        [
            "answer-evaluator",
            "pairwise",
            "queries.csv",
            "documents.csv",
            "answers.csv",
            "--verbose",
            "--data-dir",
            "tests/data/",
            "--experiment-name",
            "test",
            "--output-file",
            "test-output.json",
            "--no-save-results",
            "--force",
            "--bidirectional",
            "--add-reasoning",
        ],
    )
    assert result.exit_code == 0
    assert len(result.stdout.split("✅ Done!")) == 3
    assert len(result.stdout.split("Total evaluations: 4")) == 3
    assert result.stdout.startswith("Loaded 2 queries from")
    assert "🔎 Query ID: 0\n📜 Document ID: 0" in result.stdout
    assert "Evaluating Retrieved documents" in result.stdout
    assert "🔎 Query ID: 0\n agent2              🆚   agent1\nParsed Answer" in result.stdout
    assert "🔎 Query ID: 1\n agent1              🆚   agent2\nParsed Answer" in result.stdout
    assert os.path.exists("tests/data/test-output.json")
    os.remove("tests/data/test-output.json")



from __future__ import annotations

import asyncio
import warnings
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from textwrap import dedent
from typing import Any, Coroutine

from jinja2 import Template
from tqdm import TqdmExperimentalWarning
from tqdm.auto import tqdm
from tqdm.rich import tqdm_rich


def run(coroutine: Coroutine[Any, Any, Any]) -> Any:
    """
    Runs the given coroutine and returns its result.
    """
    return asyncio.run(coroutine)


def call_async_fn(fn: Callable[..., Coroutine[Any, Any, Any]], *args: Any, **kwargs: Any) -> Any:
    """
    Calls an asynchronous function, either in the current event loop
    or in a separate thread if no loop is running.
    """
    try:
        asyncio.get_running_loop()
        with ThreadPoolExecutor(max_workers=1) as executor:
            # Execute the coroutine using the `run` function in a thread pool
            future = executor.submit(run, fn(*args, **kwargs))
            return future.result()
    except RuntimeError:
        # If no running loop is detected, run the coroutine directly
        return asyncio.run(fn(*args, **kwargs))


def get_pbar(
    total: int, use_rich: bool = True, ncols: int = 100, desc: str = "Evaluating", disable: bool = False
) -> tqdm | tqdm_rich:
    if use_rich:
        warnings.filterwarnings("ignore", category=TqdmExperimentalWarning)
        return tqdm_rich(total=total, ncols=ncols, desc=desc, disable=disable)
    else:
        return tqdm(total=total, ncols=ncols, desc=desc, disable=disable)


def string_to_template(src: str) -> Template:
    clean_str = dedent(src).strip()
    template = Template(clean_str)
    template._ragelo_source = src  # type: ignore
    return template



"""Parse arguments for the cli app"""

from __future__ import annotations

import collections.abc
import inspect
import sys
from typing import Any, Callable, get_args, get_origin, get_type_hints

from typer.models import ArgumentInfo, OptionInfo, ParameterInfo, ParamMeta

from ragelo.types import BaseConfig

arguments = {
    "queries_csv_file",
    "documents_csv_file",
    "answers_csv_file",
    "domain_long",
}

ignore_args = {"llm_response_schema", "system_prompt", "user_prompt"}


def get_params_from_function(func: Callable[..., Any]) -> dict[str, ParamMeta]:
    if sys.version_info >= (3, 10):
        signature = inspect.signature(func, eval_str=True)
    else:
        signature = inspect.signature(func)

    type_hints = get_type_hints(func)

    params = {}

    for param in signature.parameters.values():
        annotation = param.annotation
        if param.name == "kwargs" or param.name == "args":
            continue
        if param.name in type_hints:
            annotation = type_hints[param.name]
        if inspect.isclass(annotation) and issubclass(annotation, BaseConfig):
            fields = annotation.model_fields
            for k, v in fields.items():
                if k in ignore_args:
                    continue
                description = v.description
                _type = v.annotation
                _outer_type = v.annotation
                t_args = get_args(_type)
                if get_origin(_outer_type) is list:
                    _type = _outer_type
                if get_origin(_outer_type) is type(None) or _type is type(None):
                    continue
                if get_origin(_outer_type) is dict:
                    continue

                if not isinstance(v, ParameterInfo):
                    if len(t_args) > 1:
                        # To resolve the True argument type, first remove any NoneType from the list of types"
                        _t_args = [t for t in t_args if t is not type(None)]
                        if len(_t_args) == 1:
                            _type = _t_args[0]
                        if (
                            get_origin(_outer_type) == collections.abc.Callable
                            or get_origin(_type) == collections.abc.Callable
                        ):
                            # ignore the callable type and move on.
                            continue
                        elif len(_t_args) > 1:
                            _type = _t_args[0]
                    if k in arguments:
                        argument = ArgumentInfo(default=v.default, help=description)
                        params[k] = ParamMeta(name=k, default=argument, annotation=_type)
                    else:
                        option = OptionInfo(
                            default=v.default,
                            default_factory=v.default_factory,  # type: ignore
                            help=description,
                        )
                        params[k] = ParamMeta(name=k, default=option, annotation=_type)
                else:
                    params[k] = ParamMeta(name=k, default=v, annotation=_type)

        else:
            params[param.name] = ParamMeta(name=param.name, default=param.default, annotation=annotation)
    return params



from __future__ import annotations

import os


def get_path(data_path: str | None, file_path: str, check_exists: bool = True) -> str:
    if data_path is None:
        data_path = "."
    if file_path.startswith("/"):
        abs_path = file_path
    abs_path = os.path.abspath(os.path.join(data_path, file_path))
    if check_exists:
        assert os.path.exists(abs_path), f"File {abs_path} does not exist"
    return abs_path



import logging

import typer

from ragelo import Experiment, get_answer_evaluator, get_llm_provider, get_retrieval_evaluator
from ragelo.cli.args import get_params_from_function
from ragelo.cli.utils import get_path
from ragelo.types import AnswerEvaluatorTypes
from ragelo.types.configurations.cli_configs import CLIPairwiseDomainExpertEvaluatorConfig, CLIPairwiseEvaluatorConfig
from ragelo.types.types import RetrievalEvaluatorTypes

typer.main.get_params_from_function = get_params_from_function  # type: ignore
app = typer.Typer()


@app.command()
def pairwise(config: CLIPairwiseEvaluatorConfig = CLIPairwiseEvaluatorConfig(), **kwargs):
    """An evaluator that evaluates RAG-based answers by comparing the answers of two agents to the same queries.

    example:
    >> ragelo answer-evaluator pairwise queries.csv answers.csv

    """
    logging.getLogger("ragelo").setLevel(logging.INFO)
    kwargs.pop("llm_response_schema", None)

    config = CLIPairwiseEvaluatorConfig(**kwargs)
    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    answers_file = get_path(config.data_dir, config.answers_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        answers_csv_path=answers_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    kwargs = config.model_dump()

    if config.add_reasoning:
        reasoner_evaluator = get_retrieval_evaluator(
            RetrievalEvaluatorTypes.REASONER,
            llm_provider=llm_provider,
            rich_print=config.rich_print,
            verbose=config.verbose,
        )
        reasoner_evaluator.evaluate_experiment(experiment)
        config.include_relevance_score = True
        config.include_raw_documents = False
    else:
        config.include_relevance_score = False
        config.include_raw_documents = True

    evaluator = get_answer_evaluator(AnswerEvaluatorTypes.PAIRWISE, config=config, llm_provider=llm_provider)
    evaluator.evaluate_experiment(experiment)
    experiment.save(output_file)


@app.command()
def expert_pairwise(
    config: CLIPairwiseDomainExpertEvaluatorConfig = CLIPairwiseDomainExpertEvaluatorConfig(),
    **kwargs,
):
    """
    An evaluator that evaluates RAG-based answers by comparing answers of two agents and impersonating a domain expert.
    """
    logging.getLogger("ragelo").setLevel(logging.INFO)
    kwargs.pop("llm_response_schema", None)

    config = CLIPairwiseDomainExpertEvaluatorConfig(**kwargs)

    llm_provider = get_llm_provider(config.llm_provider_name, **kwargs)

    queries_csv_file = get_path(config.data_dir, config.queries_csv_file)
    documents_file = get_path(config.data_dir, config.documents_csv_file)
    answers_file = get_path(config.data_dir, config.answers_csv_file)
    output_file = get_path(config.data_dir, config.output_file, check_exists=False) if config.output_file else None

    experiment = Experiment(
        experiment_name=config.experiment_name,
        save_path=output_file,
        queries_csv_path=queries_csv_file,
        documents_csv_path=documents_file,
        answers_csv_path=answers_file,
        verbose=config.verbose,
        clear_evaluations=config.force,
        rich_print=config.rich_print,
        cache_evaluations=config.save_results,
    )

    if config.add_reasoning:
        reasoner_evaluator = get_retrieval_evaluator(RetrievalEvaluatorTypes.REASONER, llm_provider=llm_provider)
        reasoner_evaluator.evaluate_experiment(experiment)
        config.include_relevance_score = True

    evaluator = get_answer_evaluator(AnswerEvaluatorTypes.PAIRWISE, config=config, llm_provider=llm_provider)
    evaluator.evaluate_experiment(experiment)
    experiment.save(output_file)



"""CLI for automatically evaluating Retrieval Augmented Generation (RAG) models."""

from ragelo.agent_rankers.base_agent_ranker import get_agent_ranker
from ragelo.evaluators.answer_evaluators import AnswerEvaluatorFactory, get_answer_evaluator
from ragelo.evaluators.retrieval_evaluators import RetrievalEvaluatorFactory, get_retrieval_evaluator
from ragelo.llm_providers.base_llm_provider import get_llm_provider
from ragelo.types import AgentAnswer, Document, Experiment, Query

__all__ = [
    "get_agent_ranker",
    "get_answer_evaluator",
    "get_retrieval_evaluator",
    "get_llm_provider",
    "AnswerEvaluatorFactory",
    "RetrievalEvaluatorFactory",
    "AgentAnswer",
    "Document",
    "Query",
    "Experiment",
]

__version__ = "0.1.0"
__app_name__ = "ragelo"
__author__ = "Zeta Alpha"
__email__ = "camara@zeta-alpha.com"
__description__ = "CLI for automatically evaluating Retrieval Augmented Generation (RAG) models."
